import warnings

import numpy as np
import pandas as pd
from pandas.api.types import CategoricalDtype
from sklearn.utils import Bunch
from sklearn.preprocessing import OneHotEncoder

from ..column import categorical_to_numeric, standardize
from ..io import loadarff
from ..util import safe_concat

__all__ = [
    "get_x_y",
    "load_arff_files_standardized",
    "load_aids",
    "load_breast_cancer",
    "load_flchain",
    "load_gbsg2",
    "load_whas500",
    "load_veterans_lung_cancer",
]


def _get_data_path(name):
    from importlib.resources import files

    return files(__package__) / "data" / name


def _get_x_y_survival(dataset, col_event, col_time, val_outcome):
    if col_event is None or col_time is None:
        y = None
        x_frame = dataset
    else:
        y = np.empty(
            dtype=[(col_event, bool), (col_time, np.float64)],
            shape=dataset.shape[0],
        )
        y[col_event] = (dataset[col_event] == val_outcome).values
        y[col_time] = dataset[col_time].values

        x_frame = dataset.drop([col_event, col_time], axis=1)

    return x_frame, y


def _get_x_y_other(dataset, col_label):
    if col_label is None:
        y = None
        x_frame = dataset
    else:
        y = dataset.loc[:, col_label]
        x_frame = dataset.drop(col_label, axis=1)

    return x_frame, y


def _create_event_time_df(data):
    # Convert the list of tuples into a structured NumPy array
    structured_data = np.array(data, dtype=[("event", "?"), ("time", "f8")])

    # Convert the structured array to a DataFrame
    df = pd.DataFrame(structured_data)

    # Change 'event' column to integer64, mapping False to 0 and True to 1
    df["event"] = df["event"].astype("int64")

    # Ensure 'time' column is float64
    df["time"] = df["time"].astype("float64")

    return df


def _encode_categorical_columns(df, categorical_columns=None):
    """
    Automatically identifies categorical columns and applies one-hot encoding.

    Parameters:
    - df (pd.DataFrame): The input DataFrame with mixed continuous and categorical variables.
    - categorical_columns (list): Optional list of column names to treat as categorical.

    Returns:
    - pd.DataFrame: A new DataFrame with one-hot encoded categorical columns.
    """
    # Automatically identify categorical columns if not provided
    if categorical_columns is None:
        categorical_columns = df.select_dtypes(
            include=["object", "category"]).columns.tolist()

    # Apply one-hot encoding to the identified categorical columns
    df_encoded = pd.get_dummies(df, columns=categorical_columns)

    # Convert boolean columns to integer (0 and 1)
    bool_columns = df_encoded.select_dtypes(include=["bool"]).columns.tolist()
    df_encoded[bool_columns] = df_encoded[bool_columns].astype(int)

    return df_encoded


def get_x_y(data_frame, attr_labels, pos_label=None, survival=True):
    """Split data frame into features and labels.

    Parameters
    ----------
    data_frame : pandas.DataFrame, shape = (n_samples, n_columns)
        A data frame.

    attr_labels : sequence of str or None
        A list of one or more columns that are considered the label.
        If `survival` is `True`, then attr_labels has two elements:
        1) the name of the column denoting the event indicator, and
        2) the name of the column denoting the survival time.
        If the sequence contains `None`, then labels are not retrieved
        and only a data frame with features is returned.

    pos_label : any, optional
        Which value of the event indicator column denotes that a
        patient experienced an event. This value is ignored if
        `survival` is `False`.

    survival : bool, optional, default: True
        Whether to return `y` that can be used for survival analysis.

    Returns
    -------
    X : pandas.DataFrame, shape = (n_samples, n_columns - len(attr_labels))
        Data frame containing features.

    y : None or pandas.DataFrame, shape = (n_samples, len(attr_labels))
        Data frame containing columns with supervised information.
        If `survival` was `True`, then the column denoting the event
        indicator will be boolean and survival times will be float.
        If `attr_labels` contains `None`, y is set to `None`.
    """
    if survival:
        if len(attr_labels) != 2:
            raise ValueError(
                f"expected sequence of length two for attr_labels, but got {len(attr_labels)}")
        if pos_label is None:
            raise ValueError(
                "pos_label needs to be specified if survival=True")
        return _get_x_y_survival(data_frame, attr_labels[0], attr_labels[1], pos_label)

    return _get_x_y_other(data_frame, attr_labels)


def _loadarff_with_index(filename):
    dataset = loadarff(filename)
    if "index" in dataset.columns:
        if isinstance(dataset["index"].dtype, CategoricalDtype):
            # concatenating categorical index may raise TypeError
            # see https://github.com/pandas-dev/pandas/issues/14586
            dataset["index"] = dataset["index"].astype(object)
        dataset.set_index("index", inplace=True)
    return dataset


def load_arff_files_standardized(
    path_training,
    attr_labels,
    pos_label=None,
    path_testing=None,
    survival=True,
    standardize_numeric=True,
    to_numeric=True,
):
    """Load dataset in ARFF format.

    Parameters
    ----------
    path_training : str
        Path to ARFF file containing data.

    attr_labels : sequence of str
        Names of attributes denoting dependent variables.
        If ``survival`` is set, it must be a sequence with two items:
        the name of the event indicator and the name of the survival/censoring time.

    pos_label : any type, optional
        Value corresponding to an event in survival analysis.
        Only considered if ``survival`` is ``True``.

    path_testing : str, optional
        Path to ARFF file containing hold-out data. Only columns that are available in both
        training and testing are considered (excluding dependent variables).
        If ``standardize_numeric`` is set, data is normalized by considering both training
        and testing data.

    survival : bool, optional, default: True
        Whether the dependent variables denote event indicator and survival/censoring time.

    standardize_numeric : bool, optional, default: True
        Whether to standardize data to zero mean and unit variance.
        See :func:`survivalist.column.standardize`.

    to_numeric : boo, optional, default: True
        Whether to convert categorical variables to numeric values.
        See :func:`survivalist.column.categorical_to_numeric`.

    Returns
    -------
    x_train : pandas.DataFrame, shape = (n_train, n_features)
        Training data.

    y_train : pandas.DataFrame, shape = (n_train, n_labels)
        Dependent variables of training data.

    x_test : None or pandas.DataFrame, shape = (n_train, n_features)
        Testing data if `path_testing` was provided.

    y_test : None or pandas.DataFrame, shape = (n_train, n_labels)
        Dependent variables of testing data if `path_testing` was provided.
    """
    dataset = _loadarff_with_index(path_training)

    x_train, y_train = get_x_y(dataset, attr_labels, pos_label, survival)

    if path_testing is not None:
        x_test, y_test = _load_arff_testing(
            path_testing, attr_labels, pos_label, survival)

        if len(x_train.columns.symmetric_difference(x_test.columns)) > 0:
            warnings.warn(
                "Restricting columns to intersection between training and testing data",
                stacklevel=2,
            )

            cols = x_train.columns.intersection(x_test.columns)
            if len(cols) == 0:
                raise ValueError(
                    "columns of training and test data do not intersect")

            x_train = x_train.loc[:, cols]
            x_test = x_test.loc[:, cols]

        x = safe_concat((x_train, x_test), axis=0)
        if standardize_numeric:
            x = standardize(x)
        if to_numeric:
            x = categorical_to_numeric(x)

        n_train = x_train.shape[0]
        x_train = x.iloc[:n_train, :]
        x_test = x.iloc[n_train:, :]
    else:
        if standardize_numeric:
            x_train = standardize(x_train)
        if to_numeric:
            x_train = categorical_to_numeric(x_train)

        x_test = None
        y_test = None

    return x_train, y_train, x_test, y_test


def _load_arff_testing(path_testing, attr_labels, pos_label, survival):
    test_dataset = _loadarff_with_index(path_testing)

    has_labels = pd.Index(attr_labels).isin(test_dataset.columns).all()
    if not has_labels:
        if survival:
            attr_labels = [None, None]
        else:
            attr_labels = None
    return get_x_y(test_dataset, attr_labels, pos_label, survival)


def load_whas500(*, return_X_y=True, as_frame=False):
    """Load and return the Worcester Heart Attack Study dataset

    The dataset has 500 samples and 14 features.
    The endpoint is death, which occurred for 215 patients (43.0%).

    See [1]_, [2]_ for further description.

    Parameters
    ----------
    return_X_y : bool, default=False
        If True, returns ``(data, target)`` instead of a Bunch object. See
        below for more information about the `data` and `target` object.

    as_frame : bool, default=False
        If True, the data is a pandas DataFrame including columns with
        appropriate dtypes (numeric). The target is
        a pandas DataFrame or Series depending on the number of target columns.
        If `return_X_y` is True, then (`data`, `target`) will be pandas
        DataFrames or Series as described below.

    Returns
    -------
    x : pandas.DataFrame
        The measurements for each patient.

    y : structured array with 2 fields
        *fstat*: boolean indicating whether the endpoint has been reached
        or the event time is right censored.

        *lenfol*: total length of follow-up (days from hospital admission date
        to date of last follow-up)

    References
    ----------
    .. [1] https://web.archive.org/web/20170114043458/http://www.umass.edu/statdata/statdata/data/

    .. [2] Hosmer, D., Lemeshow, S., May, S.:
        "Applied Survival Analysis: Regression Modeling of Time to Event Data."
        John Wiley & Sons, Inc. (2008)
    """
    fn = _get_data_path("whas500.arff")
    X, y = get_x_y(loadarff(fn), attr_labels=[
                   "fstat", "lenfol"], pos_label="1")
    if as_frame:
        return_X_y = False
    if return_X_y:
        return X, y
    y = _create_event_time_df(y)
    frame = pd.concat([_encode_categorical_columns(X), y], axis=1)
    target_names = ["event", "time"]
    fdescr = "The Worcester Heart Attack Study dataset"
    feature_names = X.columns
    return Bunch(
        data=X,
        target=y,
        frame=frame,
        target_names=target_names,
        DESCR=fdescr,
        feature_names=feature_names,
        filename=fn,
    )


def load_gbsg2(*, return_X_y=True, as_frame=False):
    """Load and return the German Breast Cancer Study Group 2 dataset

    The dataset has 686 samples and 8 features.
    The endpoint is recurrence free survival, which occurred for 299 patients (43.6%).

    See [1]_, [2]_ for further description.

    Parameters
    ----------
    return_X_y : bool, default=False
        If True, returns ``(data, target)`` instead of a Bunch object. See
        below for more information about the `data` and `target` object.

    as_frame : bool, default=False
        If True, the data is a pandas DataFrame including columns with
        appropriate dtypes (numeric). The target is
        a pandas DataFrame or Series depending on the number of target columns.
        If `return_X_y` is True, then (`data`, `target`) will be pandas
        DataFrames or Series as described below.

    Returns
    -------
    x : pandas.DataFrame
        The measurements for each patient.

    y : structured array with 2 fields
        *cens*: boolean indicating whether the endpoint has been reached
        or the event time is right censored.

        *time*: total length of follow-up

    References
    ----------
    .. [1] http://ascopubs.org/doi/abs/10.1200/jco.1994.12.10.2086

    .. [2] Schumacher, M., Basert, G., Bojar, H., et al.
        "Randomized 2 × 2 trial evaluating hormonal treatment and the duration of chemotherapy
        in node-positive breast cancer patients."
        Journal of Clinical Oncology 12, 2086–2093. (1994)
    """
    fn = _get_data_path("GBSG2.arff")
    X, y = get_x_y(loadarff(fn), attr_labels=["cens", "time"], pos_label="1")
    if as_frame:
        return_X_y = False
    if return_X_y:
        return X, y
    y = _create_event_time_df(y)

    frame = pd.concat([_encode_categorical_columns(X), y], axis=1)
    target_names = ["event", "time"]
    fdescr = "The German Breast Cancer Study Group 2 dataset"
    feature_names = X.columns
    return Bunch(
        data=X,
        target=y,
        frame=frame,
        target_names=target_names,
        DESCR=fdescr,
        feature_names=feature_names,
        filename=fn,
    )


def load_veterans_lung_cancer(*, return_X_y=True, as_frame=False):
    """Load and return data from the Veterans' Administration
    Lung Cancer Trial

    The dataset has 137 samples and 6 features.
    The endpoint is death, which occurred for 128 patients (93.4%).

    See [1]_ for further description.

    Parameters
    ----------
    return_X_y : bool, default=False
        If True, returns ``(data, target)`` instead of a Bunch object. See
        below for more information about the `data` and `target` object.

    as_frame : bool, default=False
        If True, the data is a pandas DataFrame including columns with
        appropriate dtypes (numeric). The target is
        a pandas DataFrame or Series depending on the number of target columns.
        If `return_X_y` is True, then (`data`, `target`) will be pandas
        DataFrames or Series as described below.

    Returns
    -------
    x : pandas.DataFrame
        The measurements for each patient.

    y : structured array with 2 fields
        *Status*: boolean indicating whether the endpoint has been reached
        or the event time is right censored.

        *Survival_in_days*: total length of follow-up

    References
    ----------
    .. [1] Kalbfleisch, J.D., Prentice, R.L.:
        "The Statistical Analysis of Failure Time Data." John Wiley & Sons, Inc. (2002)
    """
    fn = _get_data_path("veteran.arff")
    X, y = get_x_y(
        loadarff(fn),
        attr_labels=["Status", "Survival_in_days"],
        pos_label="dead",
    )
    if as_frame:
        return_X_y = False
    if return_X_y:
        return X, y
    y = _create_event_time_df(y)

    frame = pd.concat([_encode_categorical_columns(X), y], axis=1)
    target_names = ["event", "time"]
    fdescr = "The Veterans' Administration Lung Cancer Trial dataset"
    feature_names = X.columns
    return Bunch(
        data=X,
        target=y,
        frame=frame,
        target_names=target_names,
        DESCR=fdescr,
        feature_names=feature_names,
        filename=fn,
    )


def load_aids(endpoint="aids", return_X_y=True, as_frame=False):
    """Load and return the AIDS Clinical Trial dataset

    The dataset has 1,151 samples and 11 features.
    The dataset has 2 endpoints:

    1. AIDS defining event, which occurred for 96 patients (8.3%)
    2. Death, which occurred for 26 patients (2.3%)

    See [1]_, [2]_ for further description.

    Parameters
    ----------
    endpoint : aids|death
        The endpoint

    return_X_y : bool, default=False
        If True, returns ``(data, target)`` instead of a Bunch object. See
        below for more information about the `data` and `target` object.

    as_frame : bool, default=False
        If True, the data is a pandas DataFrame including columns with
        appropriate dtypes (numeric). The target is
        a pandas DataFrame or Series depending on the number of target columns.
        If `return_X_y` is True, then (`data`, `target`) will be pandas
        DataFrames or Series as described below.

    Returns
    -------
    x : pandas.DataFrame
        The measurements for each patient.

    y : structured array with 2 fields
        *censor*: boolean indicating whether the endpoint has been reached
        or the event time is right censored.

        *time*: total length of follow-up

        If ``endpoint`` is death, the fields are named *censor_d* and *time_d*.

    References
    ----------
    .. [1] https://web.archive.org/web/20170114043458/http://www.umass.edu/statdata/statdata/data/

    .. [2] Hosmer, D., Lemeshow, S., May, S.:
        "Applied Survival Analysis: Regression Modeling of Time to Event Data."
        John Wiley & Sons, Inc. (2008)
    """
    labels_aids = ["censor", "time"]
    labels_death = ["censor_d", "time_d"]
    if endpoint == "aids":
        attr_labels = labels_aids
        drop_columns = labels_death
    elif endpoint == "death":
        attr_labels = labels_death
        drop_columns = labels_aids
    else:
        raise ValueError("endpoint must be 'aids' or 'death'")

    fn = _get_data_path("actg320.arff")
    x, y = get_x_y(loadarff(fn), attr_labels=attr_labels, pos_label="1")
    x.drop(drop_columns, axis=1, inplace=True)
    if as_frame:
        return_X_y = False
    if return_X_y:
        return x, y
    y = _create_event_time_df(y)
    frame = pd.concat([x, y], axis=1)
    target_names = ["event", "time"]
    fdescr = "The AIDS Clinical Trial dataset"
    feature_names = x.columns
    return Bunch(
        data=x,
        target=y,
        frame=frame,
        target_names=target_names,
        DESCR=fdescr,
        feature_names=feature_names,
        filename=fn,
    )


def load_breast_cancer(*, return_X_y=True, as_frame=False):
    """Load and return the breast cancer dataset

    The dataset has 198 samples and 80 features.
    The endpoint is the presence of distance metastases, which occurred for 51 patients (25.8%).

    See [1]_, [2]_ for further description.

    Parameters
    ----------
    return_X_y : bool, default=False
        If True, returns ``(data, target)`` instead of a Bunch object. See
        below for more information about the `data` and `target` object.

    as_frame : bool, default=False
        If True, the data is a pandas DataFrame including columns with
        appropriate dtypes (numeric). The target is
        a pandas DataFrame or Series depending on the number of target columns.
        If `return_X_y` is True, then (`data`, `target`) will be pandas
        DataFrames or Series as described below.

    Returns
    -------
    x : pandas.DataFrame
        The measurements for each patient.

    y : structured array with 2 fields
        *e.tdm*: boolean indicating whether the endpoint has been reached
        or the event time is right censored.

        *t.tdm*: time to distant metastasis (days)

    References
    ----------
    .. [1] https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE7390

    .. [2] Desmedt, C., Piette, F., Loi et al.:
        "Strong Time Dependence of the 76-Gene Prognostic Signature for Node-Negative Breast Cancer
        Patients in the TRANSBIG Multicenter Independent Validation Series."
        Clin. Cancer Res. 13(11), 3207–14 (2007)
    """
    fn = _get_data_path("breast_cancer_GSE7390-metastasis.arff")
    X, y = get_x_y(loadarff(fn), attr_labels=["e.tdm", "t.tdm"], pos_label="1")
    if as_frame:
        return_X_y = False
    if return_X_y:
        return X, y
    y = _create_event_time_df(y)

    frame = pd.concat([_encode_categorical_columns(X), y], axis=1)
    target_names = ["event", "time"]
    fdescr = "The breast cancer dataset"
    feature_names = X.columns
    return Bunch(
        data=X,
        target=y,
        frame=frame,
        target_names=target_names,
        DESCR=fdescr,
        feature_names=feature_names,
        filename=fn,
    )


def load_flchain(*, return_X_y=True, as_frame=False):
    """Load and return assay of serum free light chain for 7874 subjects.

    The dataset has 7874 samples and 9 features:

        1. age: age in years
        2. sex: F=female, M=male
        3. sample.yr: the calendar year in which a blood sample was obtained
        4. kappa: serum free light chain, kappa portion
        5. lambda: serum free light chain, lambda portion
        6. flc.grp: the serum free light chain group for the subject, as used in the original analysis
        7. creatinine: serum creatinine
        8. mgus: whether the subject had been diagnosed with monoclonal gammapothy (MGUS)
        9. chapter: for those who died, a grouping of their primary cause of death by chapter headings
           of the International Code of Diseases ICD-9

    The endpoint is death, which occurred for 2169 subjects (27.5%).

    See [1]_, [2]_ for further description.

    Parameters
    ----------
    return_X_y : bool, default=False
        If True, returns ``(data, target)`` instead of a Bunch object. See
        below for more information about the `data` and `target` object.

    as_frame : bool, default=False
        If True, the data is a pandas DataFrame including columns with
        appropriate dtypes (numeric). The target is
        a pandas DataFrame or Series depending on the number of target columns.
        If `return_X_y` is True, then (`data`, `target`) will be pandas
        DataFrames or Series as described below.

    Returns
    -------
    x : pandas.DataFrame
        The measurements for each patient.

    y : structured array with 2 fields
        *death*: boolean indicating whether the subject died
        or the event time is right censored.

        *futime*: total length of follow-up or time of death.

    References
    ----------
    .. [1] https://doi.org/10.1016/j.mayocp.2012.03.009

    .. [2] Dispenzieri, A., Katzmann, J., Kyle, R., Larson, D., Therneau, T., Colby, C., Clark, R.,
           Mead, G., Kumar, S., Melton III, LJ. and Rajkumar, SV.
           Use of nonclonal serum immunoglobulin free light chains to predict overall survival in
           the general population, Mayo Clinic Proceedings 87:512-523. (2012)
    """
    fn = _get_data_path("flchain.arff")
    X, y = get_x_y(loadarff(fn), attr_labels=[
                   "death", "futime"], pos_label="dead")
    if as_frame:
        return_X_y = False
    if return_X_y:
        return X, y
    y = _create_event_time_df(y)

    frame = pd.concat([_encode_categorical_columns(X), y], axis=1)
    target_names = ["event", "time"]
    fdescr = "The assay of serum free light chain dataset"
    feature_names = X.columns
    return Bunch(
        data=X,
        target=y,
        frame=frame,
        target_names=target_names,
        DESCR=fdescr,
        feature_names=feature_names,
        filename=fn,
    )



import numpy as np
from numpy.testing import assert_array_equal
import pytest

from survivalist.functions import StepFunction
from survivalist.testing import all_survival_estimators


@pytest.fixture()
def a_step_function():
    x = np.array([0, 1, 1.2, 1.75, 2, 2.1, 3, 3.94, 5.4, 9])
    y = np.array([11, 9, 9.12, 7.5, 7.25, 5.14, 3, 2.94, 2.4, 1.9])
    f = StepFunction(x, y)
    return f


@pytest.fixture()
def toy_data_exponential():
    rnd = np.random.RandomState(2)
    n_samples = 100
    x = rnd.randn(n_samples, 2)
    y = np.empty(n_samples, dtype=[("event", bool), ("time", float)])
    y["time"] = rnd.exponential(scale=np.exp(x[:, 0]), size=n_samples)
    y["event"] = rnd.binomial(1, 0.5, size=n_samples) == 1

    # ensure at least 2 uncensored events exist
    y["event"][:2] = True

    # mark entry with largest time as censored
    # see https://github.com/sebp/survivalist/issues/249
    idxmax = np.argmax(y["time"])
    y["event"][idxmax] = False
    return x, y


class TestStepFunction:
    @staticmethod
    def test_exact(a_step_function):
        actual = np.array([a_step_function(v) for v in a_step_function.x])
        assert_array_equal(actual, a_step_function.y)

    @staticmethod
    def test_not_exact(a_step_function):
        z = np.diff(a_step_function.x).min() / 2
        actual = np.array([a_step_function(v + z) for v in a_step_function.x[:-1]])
        assert_array_equal(actual, a_step_function.y[:-1])

    @staticmethod
    @pytest.mark.parametrize("value", [-100, 100, -np.finfo(float).eps * 8, np.finfo(float).eps * 8])
    def test_out_of_bounds(a_step_function, value):
        v = value
        if v > 0:
            v += a_step_function.domain[1]

        with pytest.raises(ValueError, match=r"x must be within \[0\.0+; 9\.0+\]"):
            a_step_function(v)

    @staticmethod
    def test_not_finite(a_step_function, non_finite_value):
        with pytest.raises(ValueError, match="x must be finite"):
            a_step_function(non_finite_value)

    @staticmethod
    def test_equal(a_step_function):
        x = np.array([0, 1, 1.2, 1.75, 2, 2.1, 3, 3.94, 5.4, 9])
        y = np.array([11, 9, 9.12, 7.5, 7.25, 5.14, 3, 2.94, 2.4, 1.9])
        other_step_function = StepFunction(x, y)

        assert a_step_function == other_step_function

        different_step_function = StepFunction(x + 1, y)
        assert a_step_function != different_step_function

        assert a_step_function != x


@pytest.mark.parametrize(
    "estimator_cls", [est for est in all_survival_estimators() if hasattr(est, "predict_cumulative_hazard_function")]
)
def test_predict_cumulative_hazard_function_range(estimator_cls, toy_data_exponential):
    x, y = toy_data_exponential

    estimator = estimator_cls()
    if "fit_baseline_model" in estimator.get_params():
        estimator.set_params(fit_baseline_model=True)
    estimator.fit(x, y)

    t_min = y["time"].min()
    t_max = y["time"].max()
    t_mid = (t_max - t_min) / 2.0

    for fn in estimator.predict_cumulative_hazard_function(x):
        v = fn(t_min)
        assert np.isfinite(v)
        assert v == 0

    for fn in estimator.predict_cumulative_hazard_function(x):
        v = fn(t_mid)
        assert np.isfinite(v)
        assert v >= 0

    t_smaller_min = t_min / 2
    for fn in estimator.predict_cumulative_hazard_function(x):
        v = fn(t_smaller_min)
        assert np.isfinite(v)
        assert v == 0

    for fn in estimator.predict_cumulative_hazard_function(x):
        v = fn(t_max)
        assert np.isfinite(v)
        assert v >= 0

    t_bigger_max = t_max + 1
    for fn in estimator.predict_cumulative_hazard_function(x):
        with pytest.raises(ValueError, match=r"x must be within \[[0-9.]+; [0-9.]+\]"):
            fn(t_bigger_max)


@pytest.mark.parametrize(
    "estimator_cls", [est for est in all_survival_estimators() if hasattr(est, "predict_survival_function")]
)
def test_predict_survival_function_range(estimator_cls, toy_data_exponential):
    x, y = toy_data_exponential

    estimator = estimator_cls()
    if "fit_baseline_model" in estimator.get_params():
        estimator.set_params(fit_baseline_model=True)
    estimator.fit(x, y)

    t_min = y["time"].min()
    t_max = y["time"].max()
    t_mid = (t_max - t_min) / 2.0

    for fn in estimator.predict_survival_function(x):
        v = fn(t_min)
        assert np.isfinite(v)
        assert v == 1

    for fn in estimator.predict_survival_function(x):
        v = fn(t_mid)
        assert np.isfinite(v)
        assert 0 <= v <= 1

    t_smaller_min = t_min / 2
    for fn in estimator.predict_survival_function(x):
        v = fn(t_smaller_min)
        assert np.isfinite(v)
        assert v == 1

    for fn in estimator.predict_survival_function(x):
        v = fn(t_max)
        assert np.isfinite(v)
        assert 0 <= v <= 1

    t_bigger_max = t_max + 1
    for fn in estimator.predict_survival_function(x):
        with pytest.raises(ValueError, match=r"x must be within \[[0-9.]+; [0-9.]+\]"):
            fn(t_bigger_max)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
cimport cython
from libc cimport math

import numpy as np
cimport numpy as cnp

cnp.import_array()


@cython.wraparound(False)
@cython.cdivision(True)
@cython.boundscheck(False)
def coxph_negative_gradient(cnp.npy_uint8[:] event,
                            cnp.npy_double[:] time,
                            cnp.npy_double[:] y_pred):
    cdef cnp.npy_double s
    cdef int i
    cdef int j
    cdef cnp.npy_intp n_samples = event.shape[0]

    cdef cnp.ndarray[cnp.npy_double, ndim=1] gradient = cnp.PyArray_EMPTY(1, &n_samples, cnp.NPY_DOUBLE, 0)
    cdef cnp.npy_double[:] exp_tsj = cnp.PyArray_ZEROS(1, &n_samples, cnp.NPY_DOUBLE, 0)

    cdef cnp.npy_double[:] exp_pred = np.exp(y_pred)
    with nogil:
        for i in range(n_samples):
            for j in range(n_samples):
                if time[j] >= time[i]:
                    exp_tsj[i] += exp_pred[j]

        for i in range(n_samples):
            s = 0
            for j in range(n_samples):
                if event[j] and time[i] >= time[j]:
                    s += exp_pred[i] / exp_tsj[j]
            gradient[i] = event[i] - s

    return gradient


@cython.wraparound(False)
@cython.cdivision(True)
@cython.boundscheck(False)
def coxph_loss(cnp.npy_uint8[:] event,
               cnp.npy_double[:] time,
               cnp.npy_double[:] y_pred):
    cdef cnp.npy_intp n_samples = event.shape[0]
    cdef cnp.npy_double at_risk
    cdef cnp.npy_double loss = 0
    cdef int i
    cdef int j

    with nogil:
        for i in range(n_samples):
            at_risk = 0
            for j in range(n_samples):
                if time[j] >= time[i]:
                    at_risk += math.exp(y_pred[j])
            loss += event[i] * (y_pred[i] - math.log(at_risk))

    return - loss



from .ensemble_selection import (
    EnsembleSelection,
    EnsembleSelectionRegressor,
    MeanEstimator,
)
from .stacking import Stacking

__all__ = [
    "EnsembleSelection",
    "EnsembleSelectionRegressor",
    "MeanEstimator",
    "Stacking",
]



import re

import survivalist


def test_show_versions(capsys):
    survivalist.show_versions()
    captured = capsys.readouterr()

    assert "SYSTEM" in captured.out
    assert "DEPENDENCIES" in captured.out

    # check required dependency
    assert re.search(r"numpy\s*:\s([0-9\.\+a-f]|dev)+\n", captured.out)
    assert re.search(r"pandas\s*:\s([0-9\.\+a-f]|dev)+\n", captured.out)
    assert re.search(r"scikit-learn\s*:\s([0-9\.\+a-f]|dev|post\d)+\n", captured.out)

    # check optional dependency
    assert re.search(r"matplotlib\s*:\s([0-9\.]+|None)\n", captured.out)



from collections import OrderedDict

import numpy as np
from numpy.testing import assert_array_equal
import pandas as pd
import pandas.testing as tm
import pytest

from survivalist.preprocessing import OneHotEncoder


def _encoded_data(data):
    expected = []
    for nam, col in data.items():
        if hasattr(col, "cat"):
            for cat in col.cat.categories[1:]:
                name = f"{nam}={cat}"
                s = pd.Series(col == cat, dtype=np.float64)
                expected.append((name, s))
        else:
            expected.append((nam, col))

    expected_data = pd.DataFrame.from_dict(OrderedDict(expected))
    return expected_data


@pytest.fixture()
def create_data():
    def _create_data(n_samples=117):
        rnd = np.random.RandomState(51365192)
        data_num = pd.DataFrame(rnd.rand(n_samples, 5), columns=[f"N{i}" for i in range(5)])

        dat_cat = pd.DataFrame(
            OrderedDict(
                [
                    ("binary_1", pd.Categorical.from_codes(rnd.binomial(1, 0.6, n_samples), ["Yes", "No"])),
                    ("binary_2", pd.Categorical.from_codes(rnd.binomial(1, 0.376, n_samples), ["East", "West"])),
                    ("trinary", pd.Categorical.from_codes(rnd.binomial(2, 0.76, n_samples), ["Green", "Blue", "Red"])),
                    (
                        "many",
                        pd.Categorical.from_codes(
                            rnd.binomial(5, 0.47, n_samples), ["One", "Two", "Three", "Four", "Five", "Six"]
                        ),
                    ),
                ]
            )
        )
        data = pd.concat((data_num, dat_cat), axis=1)
        return data, _encoded_data(data)

    return _create_data


class TestOneHotEncoder:
    @staticmethod
    def test_fit(create_data):
        data, expected_data = create_data()

        t = OneHotEncoder().fit(data)

        assert t.feature_names_.tolist() == ["binary_1", "binary_2", "trinary", "many"]
        assert set(t.encoded_columns_) == set(expected_data.columns)

        assert t.categories_ == {k: data[k].cat.categories for k in ["binary_1", "binary_2", "trinary", "many"]}

    @staticmethod
    def test_fit_transform(create_data):
        data, expected_data = create_data()

        actual_data = OneHotEncoder().fit_transform(data)
        tm.assert_frame_equal(actual_data, expected_data)

    @staticmethod
    def test_transform(create_data):
        data, _ = create_data()

        t = OneHotEncoder().fit(data)
        data, expected_data = create_data(165)
        actual_data = t.transform(data)
        tm.assert_frame_equal(actual_data, expected_data)

        data = pd.concat((data.iloc[:, :2], data.iloc[:, 5:], data.iloc[:, 2:5]), axis=1)
        actual_data = t.transform(data)
        tm.assert_frame_equal(actual_data, expected_data)

    @staticmethod
    def test_get_feature_names_out(create_data):
        data, expected_data = create_data()

        t = OneHotEncoder()
        t.fit(data)

        out_names = t.get_feature_names_out()
        assert_array_equal(out_names, expected_data.columns.values)

    @staticmethod
    def test_get_feature_names_out_shuffled(create_data):
        data, _ = create_data()
        order = np.array(["binary_1", "N0", "N3", "trinary", "binary_2", "N1", "N2", "many"])
        expected_columns = np.array(
            [
                "binary_1=No",
                "N0",
                "N3",
                "trinary=Blue",
                "trinary=Red",
                "binary_2=West",
                "N1",
                "N2",
                "many=Two",
                "many=Three",
                "many=Four",
                "many=Five",
                "many=Six",
            ]
        )

        t = OneHotEncoder()
        t.fit(data.loc[:, order])

        out_names = t.get_feature_names_out()
        assert_array_equal(out_names, expected_columns)

        with pytest.raises(ValueError, match="input_features is not equal to feature_names_in_"):
            t.get_feature_names_out(data.columns.tolist())

    @staticmethod
    def test_transform_other_columns(create_data):
        data, _ = create_data()

        t = OneHotEncoder().fit(data)
        data, _ = create_data(125)

        data_renamed = data.rename(columns={"binary_1": "renamed_1"})
        with pytest.raises(ValueError, match=r"1 features are missing from data: \['binary_1'\]"):
            t.transform(data_renamed)

        data_dropped = data.drop("trinary", axis=1)
        error_msg = "X has 8 features, but OneHotEncoder is expecting 9 features as input"
        with pytest.raises(ValueError, match=error_msg):
            t.transform(data_dropped)

        data_renamed = data.rename(columns={"binary_1": "renamed_1", "many": "too_many"})
        with pytest.raises(ValueError, match=r"2 features are missing from data: \['binary_1', 'many'\]"):
            t.transform(data_renamed)



import pytest

from survivalist.base import SurvivalAnalysisMixin
from survivalist.testing import all_survival_estimators


@pytest.mark.parametrize("estimator_cls", all_survival_estimators())
def test_survival_analysis_base_clas(estimator_cls):
    assert hasattr(estimator_cls, "fit")
    assert hasattr(estimator_cls, "predict")
    assert hasattr(estimator_cls, "score")
    assert issubclass(estimator_cls, SurvivalAnalysisMixin)



import numpy as np
from numpy.testing import assert_array_equal
import pandas as pd
import pytest

from survivalist.datasets import load_whas500
from survivalist.testing import all_survival_estimators


@pytest.mark.parametrize("estimator_cls", all_survival_estimators())
def test_pandas_inputs(estimator_cls):
    X, y = load_whas500()
    X = X.iloc[:50]
    y = y[:50]
    X_df = X.loc[:, ["age", "bmi", "chf", "gender"]].astype(float)
    X_np = X_df.values

    estimator = estimator_cls()
    if "kernel" in estimator.get_params():
        estimator.set_params(kernel="rbf")
    estimator.fit(X_df, y)
    assert hasattr(estimator, "feature_names_in_")
    assert_array_equal(estimator.feature_names_in_, np.asarray(X_df.columns, dtype=object))
    estimator.predict(X_df)

    msg = "Feature names must be in the same order as they were in fit"
    X_bad = pd.DataFrame(X_np, columns=X_df.columns.tolist()[::-1])
    with pytest.raises(ValueError, match=msg):
        estimator.predict(X_bad)

    # warns when fitted on dataframe and transforming a ndarray
    msg = f"X does not have valid feature names, but {estimator_cls.__name__} was fitted with feature names"
    with pytest.warns(UserWarning, match=msg):
        estimator.predict(X_np)

    # warns when fitted on a ndarray and transforming dataframe
    msg = f"X has feature names, but {estimator_cls.__name__} was fitted without feature names"
    estimator.fit(X_np, y)
    with pytest.warns(UserWarning, match=msg):
        estimator.predict(X_df)



from itertools import product
from queue import LifoQueue

import numpy as np
from numpy.testing import assert_array_almost_equal, assert_array_equal
import pandas as pd
import pytest
from scipy import sparse
from sklearn.model_selection import train_test_split
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import OrdinalEncoder
from sklearn.tree._tree import TREE_UNDEFINED

from survivalist.compare import compare_survival
from survivalist.datasets import load_breast_cancer, load_veterans_lung_cancer
from survivalist.nonparametric import kaplan_meier_estimator, nelson_aalen_estimator
from survivalist.tree import ExtraSurvivalTree, SurvivalTree
from survivalist.util import Surv


@pytest.fixture()
def veterans():
    return load_veterans_lung_cancer()


@pytest.fixture()
def breast_cancer():
    X, y = load_breast_cancer()
    X["er"] = X.loc[:, "er"].map({"negative": 0, "positive": 1})
    X["grade"] = X.loc[:, "grade"].map(
        {
            "intermediate": 0,
            "poorly differentiated": 1,
            "unkown": 2,
            "well differentiated": 3,
        }
    )
    return X, y


@pytest.fixture()
def toy_data():
    rnd = np.random.RandomState(1)
    n_samples = 500
    X = np.empty((n_samples, 4), dtype=float)
    X[:, :2] = rnd.normal(scale=2, size=(n_samples, 2))
    X[:, 2] = rnd.uniform(40, 80, size=n_samples)
    X[:, 3] = rnd.binomial(1, 0.4, size=n_samples)

    time = np.zeros(n_samples, dtype=float)
    groups = [
        (X[:, 0] < 0.15) & (X[:, 2] > 66),
        (X[:, 0] < 0.15) & (X[:, 2] <= 66),
        (X[:, 0] >= 0.15) & (X[:, 0] < 0.65) & (X[:, 1] >= 0.5),
        (X[:, 0] >= 0.15) & (X[:, 0] < 0.65) & (X[:, 1] < 0.5),
        (X[:, 0] >= 0.65) & (X[:, 3] == 1),
        (X[:, 0] >= 0.65) & (X[:, 3] == 0),
    ]
    scales = [3, 1, 8, 5, 9, 7]
    for g, s in zip(groups, scales):
        assert g.sum() > 0
        time[g] = 1 + rnd.lognormal(mean=s, sigma=3, size=g.sum()).astype(int)

    event = np.ones(n_samples, dtype=bool)
    event[rnd.binomial(1, 0.333, size=n_samples).astype(bool)] = False

    y = np.fromiter(zip(event, time), dtype=[("status", bool), ("time", float)])
    return X, y


def supported_float_dtypes():
    names = (
        "float16",
        "float32",
        "float64",
        "float128",
    )
    return [i for i in names if hasattr(np, i)]


def assert_curve_almost_equal(x, y):
    jumps_x = np.diff(x) != 0
    jumps_y = np.diff(y) != 0

    pytest.approx(y[0], x[0])
    assert_array_almost_equal(x[:1], y[:1])
    assert_array_almost_equal(x[1:][jumps_x], y[1:][jumps_y])


class LogrankTreeBuilder:
    def __init__(self, max_depth=4, min_leaf=20):
        self.max_depth = max_depth
        self.min_leaf = min_leaf

    def build(self, X, y):
        val, feat, stat = self._get_best_split(X, y)
        splits = LifoQueue()
        splits.put((val, feat, stat, 0, np.arange(X.shape[0])))

        node_stats = []
        while splits.qsize() > 0:
            val, feat, stat, lvl, idx = splits.get()
            s = {"feature": feat, "threshold": val, "n_node_samples": idx.shape[0], "statistic": stat, "depth": lvl}
            node_stats.append(s)

            if val == TREE_UNDEFINED:
                continue

            left = X[idx, feat] <= val
            right = idx[~left]
            left = idx[left]

            if lvl == self.max_depth - 1:
                splits.put([TREE_UNDEFINED, TREE_UNDEFINED, -np.inf, lvl + 1, right])
                splits.put([TREE_UNDEFINED, TREE_UNDEFINED, -np.inf, lvl + 1, left])
                continue

            X_right = X[right, :]
            y_right = y[right]
            s_right = self._get_best_split(X_right, y_right)
            splits.put(list(s_right) + [lvl + 1, right])

            X_left = X[left, :]
            y_left = y[left]
            s_left = self._get_best_split(X_left, y_left)
            splits.put(list(s_left) + [lvl + 1, left])

        return pd.DataFrame.from_dict(dict(zip(range(len(node_stats)), node_stats)), orient="index")

    def _get_best_split(self, X, y):
        min_leaf = self.min_leaf
        best_val = TREE_UNDEFINED
        best_feat = TREE_UNDEFINED
        best_stat = -np.inf

        if y[y.dtype.names[0]].sum() == 0:
            return best_val, best_feat, best_stat
        for j in range(X.shape[1]):
            vals = X[:, j]
            values = np.unique(vals)
            if len(values) < 2:
                continue

            for i, v in enumerate(values[:-1]):
                t = (v + values[i + 1]) * 0.5
                groups = (vals <= t).astype(int)
                if groups.sum() >= min_leaf and (X.shape[0] - groups.sum()) >= min_leaf:
                    s, _ = compare_survival(y, groups)
                    if s > best_stat:
                        best_feat = j
                        best_val = t
                        best_stat = s
        return best_val, best_feat, best_stat


def test_tree_one_split(veterans):
    X, y = veterans
    X = X.loc[:, "Karnofsky_score"].values[:, np.newaxis]

    tree = SurvivalTree(max_depth=1)
    tree.fit(X, y)

    stats = LogrankTreeBuilder(max_depth=1).build(X, y)

    assert tree.tree_.capacity == stats.shape[0]
    assert_array_equal(tree.tree_.feature, stats.loc[:, "feature"].values)
    assert_array_equal(tree.tree_.n_node_samples, stats.loc[:, "n_node_samples"].values)
    assert_array_almost_equal(tree.tree_.threshold, stats.loc[:, "threshold"].values)

    expected_time = np.array(
        [
            1,
            2,
            3,
            4,
            7,
            8,
            10,
            11,
            12,
            13,
            15,
            16,
            18,
            19,
            20,
            21,
            22,
            24,
            25,
            27,
            29,
            30,
            31,
            33,
            35,
            36,
            42,
            43,
            44,
            45,
            48,
            49,
            51,
            52,
            53,
            54,
            56,
            59,
            61,
            63,
            72,
            73,
            80,
            82,
            83,
            84,
            87,
            90,
            92,
            95,
            97,
            99,
            100,
            103,
            105,
            110,
            111,
            112,
            117,
            118,
            122,
            123,
            126,
            132,
            133,
            139,
            140,
            143,
            144,
            151,
            153,
            156,
            162,
            164,
            177,
            182,
            186,
            200,
            201,
            216,
            228,
            231,
            242,
            250,
            260,
            278,
            283,
            287,
            314,
            340,
            357,
            378,
            384,
            389,
            392,
            411,
            467,
            553,
            587,
            991,
            999,
        ],
        dtype=float,
    )
    assert_array_equal(tree.unique_times_, expected_time)
    assert_array_equal(tree.unique_times_[~tree.is_event_time_], np.array([83, 97, 123, 182], dtype=float))

    threshold = stats.loc[0, "threshold"]
    m = X[:, 0] <= threshold
    y_left = y[m]
    _, chf_left = nelson_aalen_estimator(y_left["Status"], y_left["Survival_in_days"])

    y_right = y[~m]
    _, chf_right = nelson_aalen_estimator(y_right["Status"], y_right["Survival_in_days"])

    X_pred = np.array([[threshold - 10], [threshold + 10]])
    chf_pred = tree.predict_cumulative_hazard_function(X_pred, return_array=True)

    assert_curve_almost_equal(chf_pred[0], chf_left)
    assert_curve_almost_equal(chf_pred[1], chf_right)

    mrt_pred = tree.predict(X_pred)
    assert_array_almost_equal(mrt_pred, np.array([196.55878, 86.14939]))

    _, surv_left = kaplan_meier_estimator(y_left["Status"], y_left["Survival_in_days"])
    _, surv_right = kaplan_meier_estimator(y_right["Status"], y_right["Survival_in_days"])

    surv_pred = tree.predict_survival_function(X_pred, return_array=True)

    assert_curve_almost_equal(surv_pred[0], surv_left)
    assert_curve_almost_equal(surv_pred[1], surv_right)


def test_tree_two_split(veterans):
    X, y = veterans
    X = X.loc[:, "Karnofsky_score"].values[:, np.newaxis]

    tree = SurvivalTree(max_depth=2, max_features=1)
    tree.fit(X, y)

    assert tree.tree_.capacity == 7
    assert_array_equal(
        tree.tree_.threshold,
        np.array([45.0, 25.0, TREE_UNDEFINED, TREE_UNDEFINED, 87.5, TREE_UNDEFINED, TREE_UNDEFINED]),
    )
    expected_size = np.array([X.shape[0], 38, 8, 30, 99, 91, 8])
    assert_array_equal(tree.tree_.n_node_samples, expected_size)

    X_pred = np.array([66.05, 87.91, 45.62, 40.18, 50.65, 71.24, 96.21, 33.33, 11.57, 94.28]).reshape(-1, 1)
    mrt_pred = tree.predict(X_pred)
    expected_risk = np.array(
        [
            96.7044629620645,
            19.6309523809524,
            96.7044629620645,
            179.264571990757,
            96.7044629620645,
            96.7044629620645,
            19.6309523809524,
            179.264571990757,
            214.027380952381,
            19.6309523809524,
        ]
    )
    assert_array_almost_equal(mrt_pred, expected_risk)

    chf_pred = tree.predict_cumulative_hazard_function(X_pred, return_array=True)
    assert np.all(np.diff(chf_pred) >= 0)

    surv_pred = tree.predict_survival_function(X_pred, return_array=True)
    assert np.all(np.diff(surv_pred) <= 0)


def test_tree_split_all_censored(veterans):
    X, y = veterans
    X = X.loc[:, "Karnofsky_score"].values[:, np.newaxis]
    y["Status"][X[:, 0] > 45.0] = False

    tree = SurvivalTree(max_depth=2, max_features=1)
    tree.fit(X, y)

    assert tree.tree_.capacity == 5
    assert_array_equal(tree.tree_.threshold, np.array([45.0, 25.0, TREE_UNDEFINED, TREE_UNDEFINED, TREE_UNDEFINED]))
    expected_size = np.array([X.shape[0], 38, 8, 30, 99])
    assert_array_equal(tree.tree_.n_node_samples, expected_size)


@pytest.mark.slow()
def test_toy_data(toy_data):
    X, y = toy_data
    tree = SurvivalTree(max_depth=4, max_features=1.0, min_samples_leaf=20)
    tree.fit(X, y)

    stats = LogrankTreeBuilder(max_depth=4, min_leaf=20).build(X, y)

    assert tree.tree_.capacity == stats.shape[0]
    assert_array_equal(tree.tree_.feature, stats.loc[:, "feature"].values)
    assert_array_equal(tree.tree_.n_node_samples, stats.loc[:, "n_node_samples"].values)
    assert_array_almost_equal(tree.tree_.threshold, stats.loc[:, "threshold"].values, 5)


def test_breast_cancer_1(breast_cancer):
    X, y = breast_cancer

    tree = SurvivalTree(
        max_features="sqrt",
        max_depth=5,
        max_leaf_nodes=10,
        min_samples_split=0.06,
        min_samples_leaf=0.03,
        random_state=6,
    )
    tree.fit(X.values, y)

    assert tree.tree_.capacity == 19
    assert_array_equal(
        tree.tree_.feature,
        np.array(
            [
                61,
                29,
                5,
                TREE_UNDEFINED,
                40,
                65,
                TREE_UNDEFINED,
                10,
                12,
                4,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
                10,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
            ]
        ),
    )
    assert_array_equal(
        tree.tree_.n_node_samples,
        np.array(
            [
                198,
                170,
                28,
                8,
                20,
                164,
                6,
                59,
                105,
                74,
                31,
                9,
                65,
                13,
                7,
                39,
                20,
                7,
                13,
            ]
        ),
    )
    assert_array_almost_equal(
        tree.tree_.threshold,
        np.array(
            [
                10.97448,
                11.10251,
                11.34859,
                TREE_UNDEFINED,
                10.53533,
                8.08848,
                TREE_UNDEFINED,
                10.86403,
                10.14138,
                11.49171,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
                11.01874,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
            ]
        ),
        5,
    )


def test_breast_cancer_2(breast_cancer):
    X, y = breast_cancer

    tree = SurvivalTree(
        max_features="log2", splitter="random", max_depth=5, min_samples_split=30, min_samples_leaf=15, random_state=6
    )
    tree.fit(X.values, y)

    assert tree.tree_.capacity == 11
    assert_array_equal(
        tree.tree_.feature,
        np.array(
            [
                55,
                14,
                TREE_UNDEFINED,
                60,
                23,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
                31,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
            ]
        ),
    )
    assert_array_equal(
        tree.tree_.n_node_samples,
        np.array(
            [
                198,
                153,
                76,
                77,
                46,
                16,
                30,
                31,
                16,
                15,
                45,
            ]
        ),
    )
    assert_array_almost_equal(
        tree.tree_.threshold,
        np.array(
            [
                11.3019,
                9.0768,
                TREE_UNDEFINED,
                8.6903,
                6.83564,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
                10.66262,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
                TREE_UNDEFINED,
            ]
        ),
        5,
    )


def test_fit_int_time(breast_cancer):
    X, y = breast_cancer
    y_int = np.empty(y.shape[0], dtype=[(y.dtype.names[0], bool), (y.dtype.names[1], int)])
    y_int[:] = y

    tree_f = SurvivalTree(
        max_features="log2",
        splitter="random",
        max_depth=5,
        min_samples_split=30,
        min_samples_leaf=15,
        random_state=6,
    ).fit(X, y)

    tree_i = SurvivalTree(
        max_features="log2",
        splitter="random",
        max_depth=5,
        min_samples_split=30,
        min_samples_leaf=15,
        random_state=6,
    ).fit(X, y_int)

    assert_array_almost_equal(tree_f.unique_times_, tree_i.unique_times_)
    assert_array_equal(tree_f.tree_.feature, tree_i.tree_.feature)
    assert_array_equal(tree_f.tree_.n_node_samples, tree_i.tree_.n_node_samples)
    assert_array_almost_equal(tree_f.tree_.threshold, tree_i.tree_.threshold)


@pytest.mark.parametrize("dtype,missing", product(supported_float_dtypes(), [False, True]))
def test_fit_dtype(toy_data, dtype, missing):
    X, y = toy_data
    if missing:
        X[:23, 0] = np.nan
    X = X.astype(dtype)

    tree = SurvivalTree()
    tree.fit(X, y)
    assert hasattr(tree, "tree_")

    pred = tree.predict(X)
    assert pred.shape[0] == X.shape[0]


@pytest.mark.parametrize("func", ["predict_survival_function", "predict_cumulative_hazard_function"])
def test_predict_step_function(breast_cancer, func):
    X, y = breast_cancer

    tree = SurvivalTree(
        max_features="log2",
        splitter="random",
        max_depth=5,
        min_samples_split=30,
        min_samples_leaf=15,
        random_state=6,
    )
    tree.fit(X.iloc[10:], y[10:])

    pred_fn = getattr(tree, func)

    ret_array = pred_fn(X.iloc[:10], return_array=True)
    fn_array = pred_fn(X.iloc[:10], return_array=False)

    assert ret_array.shape[0] == fn_array.shape[0]

    for fn, arr in zip(fn_array, ret_array):
        assert_array_almost_equal(fn.x, tree.unique_times_)
        assert_array_almost_equal(fn.y, arr)


@pytest.mark.parametrize("func", ["predict_survival_function", "predict_cumulative_hazard_function"])
def test_pipeline_predict(breast_cancer, func):
    X_num, y = breast_cancer
    X_num = X_num.loc[:, ["er", "grade"]].values

    tree = SurvivalTree().fit(X_num[10:], y[10:])

    X_str, _ = load_breast_cancer()
    X_str = X_str.loc[:, ["er", "grade"]].values

    pipe = make_pipeline(OrdinalEncoder(), SurvivalTree())
    pipe.fit(X_str[10:], y[10:])

    tree_pred = getattr(tree, func)(X_num[:10], return_array=True)
    pipe_pred = getattr(pipe, func)(X_str[:10], return_array=True)

    assert_array_almost_equal(tree_pred, pipe_pred)


@pytest.mark.parametrize("n_features", [1, 3, 5, 10])
def test_predict_wrong_features(toy_data, n_features):
    X, y = toy_data
    tree = SurvivalTree(max_depth=1)
    tree.fit(X, y)

    X_new = np.random.randn(12, n_features)
    with pytest.raises(
        ValueError, match=f"X has {n_features} features, but SurvivalTree is expecting 4 features as input."
    ):
        tree.predict(X_new)


@pytest.mark.parametrize("val", [0, 0.0, -1, -1e-6, -1512])
def test_max_depth(fake_data, val):
    X, y = fake_data
    tree = SurvivalTree(max_depth=val)

    msg = (
        r"The 'max_depth' parameter of SurvivalTree must be "
        rf"an int in the range \[1, inf\) or None\. Got {val!r} instead\."
    )
    with pytest.raises(ValueError, match=msg):
        tree.fit(X, y)


@pytest.mark.parametrize("val", [0, 0.0, -1, -1e-6, -1512, 10.0, 0.51, 1.0, np.nan, np.inf])
def test_min_samples_leaf(fake_data, val):
    X, y = fake_data
    tree = SurvivalTree(min_samples_leaf=val)

    msg = (
        r"The 'min_samples_leaf' parameter of SurvivalTree must be an int in the range \[1, inf\) or "
        rf"a float in the range \(0\.0, 0\.5\]\. Got {val!r} instead\."
    )
    with pytest.raises(ValueError, match=msg):
        tree.fit(X, y)


@pytest.mark.parametrize("val", [0, 0.0, 1, 1.0, -1, -1e-6, -1512, 10.0, 1.000001, np.nan, -np.inf, np.inf])
def test_min_samples_split(fake_data, val):
    X, y = fake_data
    tree = SurvivalTree(min_samples_split=val)

    msg = (
        r"The 'min_samples_split' parameter of SurvivalTree must be an int in the range \[2, inf\) or "
        rf"a float in the range \(0\.0, 1\.0\)\. Got {val!r} instead\."
    )
    with pytest.raises(ValueError, match=msg):
        tree.fit(X, y)


@pytest.mark.parametrize("val", [1, -1, -1e-6, -1512, 0.500001, np.nan, -np.inf, np.inf])
def test_min_weight_fraction_leaf(fake_data, val):
    X, y = fake_data
    tree = SurvivalTree(min_weight_fraction_leaf=val)

    msg = (
        r"The 'min_weight_fraction_leaf' parameter of SurvivalTree must be "
        rf"a float in the range \[0\.0, 0\.5\]\. Got {val!r} instead\."
    )
    with pytest.raises(ValueError, match=msg):
        tree.fit(X, y)


@pytest.mark.parametrize("val", [0, 0.0, "", "None", "sqrt_", "log10", "car"])
def test_max_features_invalid(fake_data, val):
    X, y = fake_data
    tree = SurvivalTree(max_features=val)

    msg = (
        r"The 'max_features' parameter of SurvivalTree must be "
        r"an int in the range \[1, inf\), a float in the range \(0\.0, 1\.0\], "
        r"a str among {.+} or None\."
    )
    with pytest.raises(ValueError, match=msg):
        tree.fit(X, y)


@pytest.mark.parametrize("val", [12, 13, 100, 865411])
def test_max_features_too_large(fake_data, val):
    X, y = fake_data
    tree = SurvivalTree(max_features=val)

    with pytest.raises(ValueError, match=r"max_features must be in \(0, n_features\]"):
        tree.fit(X, y)


@pytest.mark.parametrize("val", [12.0, 13.1, 1.11, np.nan, np.inf])
def test_max_leaf_nodes_no_int(fake_data, val):
    X, y = fake_data
    tree = SurvivalTree(max_leaf_nodes=val)

    msg = r"The 'max_leaf_nodes' parameter of SurvivalTree must be an int in the range \[2, inf\) or None\."
    with pytest.raises(ValueError, match=msg):
        tree.fit(X, y)


@pytest.mark.parametrize("val", [0, 1])
def test_max_leaf_nodes_too_small(fake_data, val):
    X, y = fake_data
    tree = SurvivalTree(max_leaf_nodes=val)

    msg = r"The 'max_leaf_nodes' parameter of SurvivalTree must be an int in the range \[2, inf\) or None\."
    with pytest.raises(ValueError, match=msg):
        tree.fit(X, y)


def test_apply(veterans):
    X, y = veterans
    X = X.loc[:, "Karnofsky_score"].values[:, np.newaxis].astype(np.float32)

    tree = SurvivalTree(max_depth=2, max_features=1)
    tree.fit(X, y)

    X_trans = tree.apply(X)

    assert X_trans.shape[0] == X.shape[0]
    assert all(X_trans >= 0)
    assert all(X_trans < tree.tree_.node_count)

    X_path = tree.decision_path(X).toarray()

    assert X_path.shape[0] == X.shape[0]
    assert X_path.shape[1] == tree.tree_.node_count

    ones = X_path[np.arange(X.shape[0]), X_trans]
    assert_array_equal(ones, np.ones(X.shape[0]))


def test_apply_sparse(veterans):
    X, y = veterans
    X = X.loc[:, "Karnofsky_score"].values[:, np.newaxis].astype(np.float32)
    X_sparse = sparse.csr_matrix(X)
    tree = SurvivalTree(max_depth=2, max_features=1)
    tree.fit(X_sparse, y)

    X_trans = tree.apply(X_sparse)

    assert X_trans.shape[0] == X.shape[0]
    assert all(X_trans >= 0)
    assert all(X_trans < tree.tree_.node_count)

    X_path = tree.decision_path(X_sparse).toarray()

    assert X_path.shape[0] == X.shape[0]
    assert X_path.shape[1] == tree.tree_.node_count

    ones = X_path[np.arange(X.shape[0]), X_trans]
    assert_array_equal(ones, np.ones(X.shape[0]))


def test_predict_sparse(make_whas500):
    seed = 42
    whas500 = make_whas500(to_numeric=True)
    X, y = whas500.x, whas500.y
    # Duplicates values in whas500 leads to assert errors because of
    # tie resolution during tree fitting.
    # Using a synthetic dataset resolves this issue.
    X = np.random.RandomState(seed).binomial(n=5, p=0.1, size=X.shape)

    X_train, X_test, y_train, _ = train_test_split(X, y, random_state=seed)

    tree = SurvivalTree(min_samples_leaf=10, random_state=seed)
    tree.fit(X_train, y_train)
    y_pred = tree.predict(X_test)
    y_cum_h = tree.predict_cumulative_hazard_function(X_test)
    y_surv = tree.predict_survival_function(X_test)

    X_train_csr = sparse.csr_matrix(X_train)
    X_test_csr = sparse.csr_matrix(X_test)

    tree_csr = SurvivalTree(min_samples_leaf=10, random_state=seed)
    tree_csr.fit(X_train_csr, y_train)
    y_pred_csr = tree_csr.predict(X_test_csr)
    y_cum_h_csr = tree_csr.predict_cumulative_hazard_function(X_test_csr)
    y_surv_csr = tree_csr.predict_survival_function(X_test_csr)

    assert y_pred.shape[0] == X_test.shape[0]
    assert y_pred_csr.shape[0] == X_test.shape[0]

    assert_array_equal(y_pred, y_pred_csr)
    assert_array_equal(y_cum_h, y_cum_h_csr)
    assert_array_equal(y_surv, y_surv_csr)


def test_missing_values_best_splitter_to_max_samples(veterans):
    X, y = veterans
    X = X.loc[:, "Karnofsky_score"].values[:, np.newaxis].astype(np.float32)

    tree = SurvivalTree(max_depth=1)
    tree.fit(X, y)

    y_pred_chf = tree.predict_cumulative_hazard_function([[np.nan]], return_array=True)
    y_pred_surv = tree.predict_survival_function([[np.nan]], return_array=True)
    y_pred = np.column_stack((y_pred_chf[0], y_pred_surv[0]))

    # missing values go to node with the most samples
    assert np.argmax(tree.tree_.n_node_samples[1:]) == 1
    y_expected = tree.tree_.value[2]
    assert_array_almost_equal(y_pred, y_expected)


def test_missing_values_best_splitter_to_right():
    X = np.array([[np.nan] * 8 + list(range(7))], dtype=np.float32).T
    y = Surv.from_arrays(time=np.concatenate((np.arange(8) + 10, np.arange(6, 13))), event=np.ones(15, dtype=bool))

    tree = SurvivalTree(max_depth=2)
    tree.fit(X, y)

    y_pred_chf = tree.predict_cumulative_hazard_function([[np.nan]], return_array=True)
    y_pred_surv = tree.predict_survival_function([[np.nan]], return_array=True)
    y_pred = np.column_stack((y_pred_chf[0], y_pred_surv[0]))

    # missing values go to the right
    y_expected = tree.tree_.value[4]
    assert_array_almost_equal(y_pred, y_expected)


@pytest.mark.parametrize("is_sparse", [False, True])
def test_missing_value_random_splitter_errors(is_sparse):
    X = np.array([[3, 5, 7, 11, np.nan, 13, 17, np.nan, 19]], dtype=np.float32).T
    y = Surv.from_arrays(
        event=np.array([True, True, True, False, True, False, False, False, True]),
        time=np.array([90, 80, 70, 60, 50, 40, 30, 20, 10]),
    )

    if is_sparse:
        X = sparse.csr_matrix(X)

    tree = ExtraSurvivalTree()
    with pytest.raises(ValueError, match="Input X contains NaN"):
        tree.fit(X, y)



import pytest

from survivalist.bintrees import AVLTree, RBTree


@pytest.fixture(params=[RBTree, AVLTree])
def tree(request):
    return request.param(10)


class TestBinaryTree:
    @staticmethod
    def test_insert(tree):
        for k in (12, 34, 45, 16, 35, 57):
            tree.insert(k, k)
        assert 6 == len(tree)

    @staticmethod
    def test_count_smaller(tree):
        for k in (12, 34, 45, 16, 35, 57):
            tree.insert(k, k)

        c, _ = tree.count_smaller(12)
        assert 0 == c

        c, _ = tree.count_smaller(16)
        assert 1 == c

        c, _ = tree.count_smaller(34)
        assert 2 == c

        c, _ = tree.count_smaller(35)
        assert 3 == c

        c, _ = tree.count_smaller(45)
        assert 4 == c

        c, _ = tree.count_smaller(57)
        assert 5 == c

    @staticmethod
    def test_count_larger(tree):
        for k in (12, 34, 45, 16, 35, 57):
            tree.insert(k, k)

        c, _ = tree.count_larger(12)
        assert 5 == c

        c, _ = tree.count_larger(16)
        assert 4 == c

        c, _ = tree.count_larger(34)
        assert 3 == c

        c, _ = tree.count_larger(35)
        assert 2 == c

        c, _ = tree.count_larger(45)
        assert 1 == c

        c, _ = tree.count_larger(57)
        assert 0 == c



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from sklearn.utils import check_consistent_length

__all__ = ["StepFunction"]


class StepFunction:
    """Callable step function.

    .. math::

        f(z) = a * y_i + b,
        x_i \\leq z < x_{i + 1}

    Parameters
    ----------
    x : ndarray, shape = (n_points,)
        Values on the x axis in ascending order.

    y : ndarray, shape = (n_points,)
        Corresponding values on the y axis.

    a : float, optional, default: 1.0
        Constant to multiply by.

    b : float, optional, default: 0.0
        Constant offset term.

    domain : tuple, optional
        A tuple with two entries that sets the limits of the
        domain of the step function.
        If entry is `None`, use the first/last value of `x` as limit.
    """

    def __init__(self, x, y, *, a=1.0, b=0.0, domain=(0, None)):
        check_consistent_length(x, y)
        self.x = x
        self.y = y
        self.a = a
        self.b = b
        domain_lower = self.x[0] if domain[0] is None else domain[0]
        domain_upper = self.x[-1] if domain[1] is None else domain[1]
        self._domain = (float(domain_lower), float(domain_upper))

    @property
    def domain(self):
        """Returns the domain of the function, that means
        the range of values that the function accepts.

        Returns
        -------
        lower_limit : float
            Lower limit of domain.

        upper_limit : float
            Upper limit of domain.
        """
        return self._domain

    def __call__(self, x):
        """Evaluate step function.

        Values outside the interval specified by `self.domain`
        will raise an exception.
        Values in `x` that are in the interval `[self.domain[0]; self.x[0]]`
        get mapped to `self.y[0]`.

        Parameters
        ----------
        x : float|array-like, shape=(n_values,)
            Values to evaluate step function at.

        Returns
        -------
        y : float|array-like, shape=(n_values,)
            Values of step function at `x`.
        """
        x = np.atleast_1d(x)
        if not np.isfinite(x).all():
            raise ValueError("x must be finite")
        if np.min(x) < self._domain[0] or np.max(x) > self.domain[1]:
            raise ValueError(
                f"x must be within [{self.domain[0]:f}; {self.domain[1]:f}]")

        # x is within the domain, but we need to account for self.domain[0] <= x < self.x[0]
        x = np.clip(x, a_min=self.x[0], a_max=None)

        i = np.searchsorted(self.x, x, side="left")
        not_exact = self.x[i] != x
        i[not_exact] -= 1
        value = self.a * self.y[i] + self.b
        if value.shape[0] == 1:
            return value[0]
        return value

    def __repr__(self):
        return f"StepFunction(x={self.x!r}, y={self.y!r}, a={self.a!r}, b={self.b!r})"

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return all(self.x == other.x) and all(self.y == other.y) and self.a == other.a and self.b == other.b
        return False



import numpy as np
from numpy.testing import assert_array_almost_equal, assert_array_equal
import pytest
from scipy import sparse
from sklearn.ensemble import ExtraTreesClassifier, RandomForestClassifier
from sklearn.model_selection import train_test_split
from sklearn.pipeline import make_pipeline

from survivalist.datasets import load_breast_cancer
from survivalist.ensemble import ExtraSurvivalTrees, RandomSurvivalForest
from survivalist.preprocessing import OneHotEncoder
from survivalist.testing import assert_chf_properties, assert_cindex_almost_equal, assert_survival_function_properties
from survivalist.tree import SurvivalTree

FORESTS = [
    RandomSurvivalForest,
    ExtraSurvivalTrees,
]


@pytest.mark.parametrize(
    "forest_cls, expected_c",
    [
        (RandomSurvivalForest, (0.9009168452008676, 67703, 7446, 0, 14)),
        (ExtraSurvivalTrees, (0.8400644053813091, 63130, 12019, 0, 14)),
    ],
)
def test_fit_predict(make_whas500, forest_cls, expected_c):
    whas500 = make_whas500(to_numeric=True)

    forest = forest_cls(random_state=2)
    forest.fit(whas500.x, whas500.y)

    assert len(forest.estimators_) == 100

    pred = forest.predict(whas500.x)
    assert np.isfinite(pred).all()
    assert np.all(pred >= 0)

    assert_cindex_almost_equal(whas500.y["fstat"], whas500.y["lenfol"], pred, expected_c)


def test_fit_missing_values(make_whas500):
    whas500 = make_whas500(to_numeric=True)

    rng = np.random.RandomState(42)
    mask = rng.binomial(n=1, p=0.15, size=whas500.x.shape)
    mask = mask.astype(bool)
    X = whas500.x.copy()
    X[mask] = np.nan

    X_train, y_train = X[:400], whas500.y[:400]
    X_test, y_test = X[400:], whas500.y[400:]

    forest = RandomSurvivalForest(random_state=42)
    forest.fit(X_train, y_train)

    tags = forest._get_tags()
    assert tags["allow_nan"]

    cindex = forest.score(X_test, y_test)
    assert cindex == pytest.approx(0.7444120505344995)


def test_fit_missing_values_not_supported(make_whas500):
    whas500 = make_whas500(to_numeric=True)

    rng = np.random.RandomState(42)
    mask = rng.binomial(n=1, p=0.15, size=whas500.x.shape)
    mask = mask.astype(bool)
    X = whas500.x.copy()
    X[mask] = np.nan

    forest = ExtraSurvivalTrees(random_state=42)
    with pytest.raises(ValueError, match="Input X contains NaN"):
        forest.fit(X, whas500.y)

    tags = forest._get_tags()
    assert not tags["allow_nan"]


@pytest.mark.parametrize("forst_cls,allows_nan", [(ExtraTreesClassifier, False), (RandomForestClassifier, True)])
def test_sklearn_random_forest_tags(forst_cls, allows_nan):
    est = forst_cls()

    # https://scikit-learn.org/stable/developers/develop.html#estimator-tags
    tags = est._get_tags()
    assert tags["multioutput"]
    assert tags["requires_fit"]
    assert tags["requires_y"]
    assert tags["allow_nan"] is allows_nan


@pytest.mark.parametrize("forest_cls", FORESTS)
def test_fit_int_time(make_whas500, forest_cls):
    whas500 = make_whas500(to_numeric=True)
    y = whas500.y
    y_int = np.empty(y.shape[0], dtype=[(y.dtype.names[0], bool), (y.dtype.names[1], int)])
    y_int[:] = y

    forest_f = forest_cls(oob_score=True, random_state=2).fit(whas500.x[50:], y[50:])
    forest_i = forest_cls(oob_score=True, random_state=2).fit(whas500.x[50:], y_int[50:])

    assert len(forest_f.estimators_) == len(forest_i.estimators_)
    assert forest_f.n_features_in_ == forest_i.n_features_in_
    assert forest_f.oob_score_ == forest_i.oob_score_
    assert_array_almost_equal(forest_f.unique_times_, forest_i.unique_times_)

    pred_f = forest_f.predict(whas500.x[:50])
    pred_i = forest_i.predict(whas500.x[:50])

    assert_array_almost_equal(pred_f, pred_i)


@pytest.mark.parametrize("forest_cls", FORESTS)
def test_fit_predict_chf(make_whas500, forest_cls):
    whas500 = make_whas500(to_numeric=True)

    forest = forest_cls(n_estimators=10, random_state=2)
    forest.fit(whas500.x, whas500.y)

    assert len(forest.estimators_) == 10

    chf = forest.predict_cumulative_hazard_function(whas500.x, return_array=True)
    assert chf.shape == (500, forest.unique_times_.shape[0])

    assert_chf_properties(chf)


@pytest.mark.parametrize("forest_cls", FORESTS)
def test_fit_predict_surv(make_whas500, forest_cls):
    whas500 = make_whas500(to_numeric=True)

    forest = forest_cls(n_estimators=10, random_state=2)
    forest.fit(whas500.x, whas500.y)

    assert len(forest.estimators_) == 10

    surv = forest.predict_survival_function(whas500.x, return_array=True)
    assert surv.shape == (500, forest.unique_times_.shape[0])

    assert_survival_function_properties(surv)


@pytest.mark.parametrize(
    "forest_cls, expected_oob_score", [(RandomSurvivalForest, 0.758732651), (ExtraSurvivalTrees, 0.751427165)]
)
def test_oob_score(make_whas500, forest_cls, expected_oob_score):
    whas500 = make_whas500(to_numeric=True)

    forest = forest_cls(oob_score=True, bootstrap=False, random_state=2)
    with pytest.raises(ValueError, match="Out of bag estimation only available if bootstrap=True"):
        forest.fit(whas500.x, whas500.y)

    forest.set_params(bootstrap=True)
    forest.fit(whas500.x, whas500.y)

    assert forest.oob_prediction_.shape == (whas500.x.shape[0],)
    assert forest.oob_score_ == pytest.approx(expected_oob_score)


@pytest.mark.parametrize("forest_cls", FORESTS)
@pytest.mark.parametrize("func", ["predict_survival_function", "predict_cumulative_hazard_function"])
def test_predict_step_function(make_whas500, forest_cls, func):
    whas500 = make_whas500(to_numeric=True)

    forest = forest_cls(n_estimators=10, random_state=2)
    forest.fit(whas500.x[10:], whas500.y[10:])

    pred_fn = getattr(forest, func)

    ret_array = pred_fn(whas500.x[:10], return_array=True)
    fn_array = pred_fn(whas500.x[:10], return_array=False)

    assert ret_array.shape[0] == fn_array.shape[0]

    for fn, arr in zip(fn_array, ret_array):
        assert_array_almost_equal(fn.x, forest.unique_times_)
        assert_array_almost_equal(fn.y, arr)


@pytest.mark.parametrize("forest_cls", FORESTS)
def test_oob_too_little_estimators(make_whas500, forest_cls):
    whas500 = make_whas500(to_numeric=True)

    forest = forest_cls(n_estimators=3, oob_score=True, random_state=2)
    with pytest.warns(
        UserWarning,
        match="Some inputs do not have OOB scores. "
        "This probably means too few trees were used "
        "to compute any reliable oob estimates.",
    ):
        forest.fit(whas500.x, whas500.y)


def test_fit_no_bootstrap(make_whas500):
    whas500 = make_whas500(to_numeric=True)

    forest = RandomSurvivalForest(n_estimators=10, bootstrap=False, random_state=2)
    forest.fit(whas500.x, whas500.y)

    pred = forest.predict(whas500.x)

    expected_c = (0.931881994437717, 70030, 5119, 0, 14)
    assert_cindex_almost_equal(whas500.y["fstat"], whas500.y["lenfol"], pred, expected_c)


@pytest.mark.parametrize("forest_cls", FORESTS)
def test_fit_warm_start(make_whas500, forest_cls):
    whas500 = make_whas500(to_numeric=True)

    forest = forest_cls(n_estimators=11, max_depth=2, random_state=2)
    forest.fit(whas500.x, whas500.y)

    assert len(forest.estimators_) == 11
    assert all(e.max_depth == 2 for e in forest.estimators_)

    forest.set_params(warm_start=True)
    with pytest.warns(UserWarning, match="Warm-start fitting without increasing n_estimators does not fit new trees."):
        forest.fit(whas500.x, whas500.y)

    forest.set_params(n_estimators=3)
    with pytest.raises(
        ValueError, match=r"n_estimators=3 must be larger or equal to len\(estimators_\)=11 when warm_start==True"
    ):
        forest.fit(whas500.x, whas500.y)

    forest.set_params(n_estimators=23)
    forest.fit(whas500.x, whas500.y)

    assert len(forest.estimators_) == 23
    assert all(e.max_depth == 2 for e in forest.estimators_)


@pytest.mark.parametrize("forest_cls", FORESTS)
def test_fit_with_small_max_samples(make_whas500, forest_cls):
    whas500 = make_whas500(to_numeric=True)

    # First fit with no restriction on max samples
    est1 = forest_cls(n_estimators=1, random_state=1, max_samples=None)

    # Second fit with max samples restricted to just 2
    est2 = forest_cls(n_estimators=1, random_state=1, max_samples=2)

    est1.fit(whas500.x, whas500.y)
    est2.fit(whas500.x, whas500.y)

    tree1 = est1.estimators_[0].tree_
    tree2 = est2.estimators_[0].tree_

    msg = "Tree without `max_samples` restriction should have more nodes"
    assert tree1.node_count > tree2.node_count, msg


@pytest.mark.parametrize("forest_cls", FORESTS)
@pytest.mark.parametrize("func", ["predict_survival_function", "predict_cumulative_hazard_function"])
def test_pipeline_predict(breast_cancer, forest_cls, func):
    X_str, _ = load_breast_cancer()
    X_num, y = breast_cancer

    est = forest_cls(n_estimators=10, random_state=1)
    est.fit(X_num[10:], y[10:])

    pipe = make_pipeline(OneHotEncoder(), forest_cls(n_estimators=10, random_state=1))
    pipe.fit(X_str[10:], y[10:])

    tree_pred = getattr(est, func)(X_num[:10], return_array=True)
    pipe_pred = getattr(pipe, func)(X_str[:10], return_array=True)

    assert_array_almost_equal(tree_pred, pipe_pred)


@pytest.mark.parametrize("forest_cls", FORESTS)
@pytest.mark.parametrize(
    "max_samples, exc_type, exc_msg, with_prefix",
    [
        (int(1e9), ValueError, "`max_samples` must be <= n_samples=500 but got value 1000000000", False),
        (1.0 + 1e-7, ValueError, r"Got 1\.0000001 instead", True),
        (2.0, ValueError, r"Got 2\.0 instead", True),
        (0.0, ValueError, r"Got 0\.0 instead", True),
        (np.nan, ValueError, "Got nan instead", True),
        (np.inf, ValueError, r"Got inf instead", True),
        ("str max_samples?!", TypeError, r"Got 'str max_samples\?!' instead", True),
        (np.ones(2), TypeError, r"Got array\(\[1\., 1\.\]\) instead", True),
        (0, ValueError, r"Got 0 instead", True),
    ],
)
def test_fit_max_samples(make_whas500, forest_cls, max_samples, exc_type, exc_msg, with_prefix):
    whas500 = make_whas500(to_numeric=True)
    forest = forest_cls(max_samples=max_samples)
    prefix = (
        f"The 'max_samples' parameter of {forest_cls.__name__} must be None, "
        r"a float in the range \(0\.0, 1\.0] or an int in the range \[1, inf\)\. "
    )
    if with_prefix:
        msg = prefix + exc_msg
    else:
        msg = exc_msg
    with pytest.raises(exc_type, match=msg):
        forest.fit(whas500.x, whas500.y)


@pytest.mark.parametrize("forest_cls", FORESTS)
@pytest.mark.parametrize("max_features", [0, 0.0, 3.0, "", "None", "sqrt_", "log10", "car"])
def test_fit_max_features(make_whas500, forest_cls, max_features):
    whas500 = make_whas500(to_numeric=True)
    forest = forest_cls(max_features=max_features)

    msg = (
        f"The 'max_features' parameter of {forest_cls.__name__} must be "
        r"an int in the range \[1, inf\), a float in the range \(0\.0, 1\.0\], "
        r"a str among {.+} or None\."
    )
    with pytest.raises(ValueError, match=msg):
        forest.fit(whas500.x, whas500.y)


@pytest.mark.parametrize("forest_cls", FORESTS)
def test_apply(make_whas500, forest_cls):
    whas500 = make_whas500(to_numeric=True)

    forest = forest_cls()
    forest.fit(whas500.x, whas500.y)

    x_trans = forest.apply(whas500.x)

    assert x_trans.shape[0] == whas500.x.shape[0]
    assert x_trans.shape[1] == forest.n_estimators

    x_path, _ = forest.decision_path(whas500.x)

    assert x_path.toarray().shape[0] == whas500.x.shape[0]


@pytest.mark.parametrize("forest_cls", FORESTS)
def test_apply_sparse(make_whas500, forest_cls):
    whas500 = make_whas500(to_numeric=True)

    forest = forest_cls()
    X, y = whas500.x, whas500.y
    X_csr = sparse.csr_matrix(X)
    forest.fit(X_csr, y)

    X_trans = forest.apply(X_csr)

    assert X_trans.shape[0] == X.shape[0]
    assert X_trans.shape[1] == forest.n_estimators

    X_path, _ = forest.decision_path(X_csr)

    assert X_path.toarray().shape[0] == X.shape[0]


@pytest.mark.parametrize("forest_cls", FORESTS)
def test_predict_sparse(make_whas500, forest_cls):
    seed = 42
    whas500 = make_whas500(to_numeric=True)
    X, y = whas500.x, whas500.y
    X = np.random.RandomState(seed).binomial(n=5, p=0.1, size=X.shape)

    X_train, X_test, y_train, _ = train_test_split(X, y, random_state=seed)

    forest = forest_cls(random_state=seed)
    forest.fit(X_train, y_train)
    y_pred = forest.predict(X_test)
    y_cum_h = forest.predict_cumulative_hazard_function(X_test)
    y_surv = forest.predict_survival_function(X_test)

    X_train_csr = sparse.csr_matrix(X_train)
    X_test_csr = sparse.csr_matrix(X_test)

    forest_csr = forest_cls(random_state=seed)
    forest_csr.fit(X_train_csr, y_train)
    y_pred_csr = forest_csr.predict(X_test_csr)
    y_cum_h_csr = forest_csr.predict_cumulative_hazard_function(X_test_csr)
    y_surv_csr = forest_csr.predict_survival_function(X_test_csr)

    assert y_pred.shape[0] == X_test.shape[0]
    assert y_pred_csr.shape[0] == X_test.shape[0]

    assert_array_equal(y_pred, y_pred_csr)
    assert_array_equal(y_cum_h_csr, y_cum_h)
    assert_array_equal(y_surv, y_surv_csr)


@pytest.mark.parametrize(
    "est_cls,params",
    [
        (SurvivalTree, {"min_samples_leaf": 10, "random_state": 42}),
        (RandomSurvivalForest, {"n_estimators": 10, "min_samples_leaf": 10, "random_state": 42}),
        (ExtraSurvivalTrees, {"n_estimators": 10, "min_samples_leaf": 10, "random_state": 42}),
    ],
)
def test_predict_low_memory(make_whas500, est_cls, params):
    whas500 = make_whas500(to_numeric=True)
    X, y = whas500.x, whas500.y

    X_train, X_test, y_train, _ = train_test_split(X, y, random_state=params["random_state"])

    est_high = est_cls(**params)
    est_high.set_params(low_memory=False)
    est_high.fit(X_train, y_train)
    pred_high = est_high.predict(X_test)

    est_low = est_cls(**params)
    est_low.set_params(low_memory=True)
    est_low.fit(X_train, y_train)
    pred_low = est_low.predict(X_test)

    assert pred_high.shape[0] == X_test.shape[0]
    assert pred_low.shape[0] == X_test.shape[0]

    assert_array_almost_equal(pred_high, pred_low)

    msg = (
        "predict_cumulative_hazard_function is not implemented in low memory mode."
        " run fit with low_memory=False to disable low memory mode."
    )
    with pytest.raises(NotImplementedError, match=msg):
        est_low.predict_cumulative_hazard_function(X_test)

    msg = (
        "predict_survival_function is not implemented in low memory mode."
        " run fit with low_memory=False to disable low memory mode."
    )
    with pytest.raises(NotImplementedError, match=msg):
        est_low.predict_survival_function(X_test)



import numpy as np
from numpy.testing import assert_array_equal
import pytest
from sklearn.dummy import DummyRegressor
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import KFold, ParameterGrid

from survivalist.ensemble import ComponentwiseGradientBoostingSurvivalAnalysis
from survivalist.kernels import ClinicalKernelTransform
from survivalist.linear_model import IPCRidge
from survivalist.meta import EnsembleSelection, EnsembleSelectionRegressor
from survivalist.metrics import concordance_index_censored
from survivalist.svm import FastKernelSurvivalSVM, FastSurvivalSVM
from survivalist.testing import FixtureParameterFactory, assert_cindex_almost_equal
from survivalist.tree import SurvivalTree
from survivalist.util import check_array_survival


def score_cindex(est, X_test, y_test, **predict_params):
    prediction = est.predict(X_test, **predict_params)

    res = concordance_index_censored(y_test["fstat"], y_test["lenfol"], prediction)
    return res[0]


def _create_survival_ensemble(**kwargs):
    boosting_grid = ParameterGrid({"n_estimators": [100, 250], "subsample": [1.0, 0.75, 0.5]})
    alphas = np.exp(np.linspace(np.log(0.001), np.log(2), 5))
    svm_grid = ParameterGrid({"alpha": alphas})

    base_estimators = []
    for i, params in enumerate(boosting_grid):
        model = ComponentwiseGradientBoostingSurvivalAnalysis(random_state=0, **params)
        base_estimators.append((f"gbm_{i}", model))

    for i, params in enumerate(svm_grid):
        model = FastSurvivalSVM(max_iter=256, tol=5e-6, random_state=0, **params)
        base_estimators.append((f"svm_{i}", model))

    cv = KFold(n_splits=3, shuffle=True, random_state=0)
    meta = EnsembleSelection(base_estimators, n_estimators=0.4, scorer=score_cindex, cv=cv, **kwargs)
    return meta


class TestEnsembleSelectionSurvivalAnalysis:
    @staticmethod
    @pytest.mark.slow()
    def test_fit(make_whas500):
        whas500 = make_whas500(with_mean=False, with_std=False, to_numeric=True)
        meta = _create_survival_ensemble()
        assert len(meta) == 0

        meta.fit(whas500.x, whas500.y)
        assert len(meta) == 11
        assert meta.scores_.shape == (11,)

        p = meta.predict(whas500.x)

        assert_cindex_almost_equal(whas500.y["fstat"], whas500.y["lenfol"], p, (0.7863312, 59088, 16053, 8, 14))

        c_index = meta.score(whas500.x, whas500.y)
        assert c_index == pytest.approx(0.7863312)

    @staticmethod
    @pytest.mark.slow()
    def test_fit_spearman_correlation(make_whas500):
        whas500 = make_whas500(with_mean=False, with_std=False, to_numeric=True)
        meta = _create_survival_ensemble(correlation="spearman")
        assert len(meta) == 0

        meta.fit(whas500.x, whas500.y)

        p = meta.predict(whas500.x)

        assert_cindex_almost_equal(whas500.y["fstat"], whas500.y["lenfol"], p, (0.7863312, 59088, 16053, 8, 14))

    @staticmethod
    @pytest.mark.slow()
    def test_fit_kendall_correlation(make_whas500):
        whas500 = make_whas500(with_mean=False, with_std=False, to_numeric=True)
        meta = _create_survival_ensemble(correlation="kendall")
        assert len(meta) == 0

        meta.fit(whas500.x, whas500.y)

        p = meta.predict(whas500.x)

        assert_cindex_almost_equal(whas500.y["fstat"], whas500.y["lenfol"], p, (0.7663043, 57570, 17545, 34, 14))

    @staticmethod
    @pytest.mark.slow()
    def test_fit_custom_kernel(make_whas500):
        whas500 = make_whas500(with_mean=False, with_std=False, to_numeric=True)
        alphas = np.exp(np.linspace(np.log(0.001), np.log(0.5), 5))
        svm_grid = ParameterGrid({"alpha": alphas})

        transform = ClinicalKernelTransform(fit_once=True)
        transform.prepare(whas500.x_data_frame)

        base_estimators = []
        for i, params in enumerate(svm_grid):
            model = FastSurvivalSVM(max_iter=100, random_state=0, **params)
            base_estimators.append((f"svm_linear_{i}", model))

        for i, params in enumerate(svm_grid):
            model = FastKernelSurvivalSVM(
                kernel=transform.pairwise_kernel, max_iter=45, tol=1e-5, random_state=0, **params
            )
            base_estimators.append((f"svm_kernel_{i}", model))

        cv = KFold(n_splits=3, shuffle=True, random_state=0)
        meta = EnsembleSelection(base_estimators, n_estimators=0.4, scorer=score_cindex, cv=cv, n_jobs=4)

        meta.fit(whas500.x, whas500.y)
        assert len(meta) == 10
        assert meta.scores_.shape == (10,)

        p = meta.predict(whas500.x)

        assert_cindex_almost_equal(whas500.y["fstat"], whas500.y["lenfol"], p, (0.7978084, 59938, 15178, 33, 14))

    @staticmethod
    def test_feature_names_in(make_whas500):
        whas500 = make_whas500(with_mean=False, with_std=False, to_numeric=True)

        base_estimators = [(f"tree{i}", SurvivalTree(max_depth=1, max_features=1, random_state=i)) for i in range(10)]

        cv = KFold(n_splits=3, shuffle=True, random_state=0)
        meta = EnsembleSelection(base_estimators, n_estimators=0.5, scorer=score_cindex, cv=cv)

        meta.fit(whas500.x_data_frame, whas500.y)
        feature_names = whas500.x_data_frame.columns.values
        assert meta.n_features_in_ == len(feature_names)
        assert_array_equal(meta.feature_names_in_, feature_names)

        warn_msg = "X does not have valid feature names, but SurvivalTree was fitted with feature names"
        with pytest.warns(UserWarning, match=warn_msg):
            meta.predict(whas500.x[:10])

        meta.fit(whas500.x, whas500.y)
        assert meta.n_features_in_ == len(feature_names)

        with pytest.raises(AttributeError, match="object has no attribute 'feature_names_in_'"):
            meta.feature_names_in_  # pylint: disable=pointless-statement

        warn_msg = "X has feature names, but SurvivalTree was fitted without feature names"
        with pytest.warns(UserWarning, match=warn_msg):
            meta.predict(whas500.x_data_frame.iloc[:10])


class EnsembleSelectionFailureCases(FixtureParameterFactory):
    _msg_min_correlation = (
        "The 'min_correlation' parameter of EnsembleSelection must be "
        r"a float in the range \[-1\.0, 1\.0\]\. "
        "Got {value} instead\\."
    )
    _msg_correlation = (
        r"The 'correlation' parameter of EnsembleSelection must be a str among {{.+}}\. Got {value!r} instead\."
    )
    _msg_scorer = r"The 'scorer' parameter of EnsembleSelection must be a callable\. Got {value} instead\."

    def data_min_score(self):
        params = {"scorer": score_cindex, "min_score": 1.0, "cv": 3}
        match = "no base estimator exceeds min_score, try decreasing it"
        return params, ValueError, match

    def data_min_correlation_0(self):
        params = {"scorer": score_cindex, "min_correlation": 1.2}
        match = self._msg_min_correlation.format(value=1.2)
        return params, ValueError, match

    def data_min_correlation_1(self):
        params = {"scorer": score_cindex, "min_correlation": -2.1}
        match = self._msg_min_correlation.format(value=-2.1)
        return params, ValueError, match

    def data_min_correlation_2(self):
        params = {"scorer": score_cindex, "min_correlation": np.nan}
        match = self._msg_min_correlation.format(value=np.nan)
        return params, ValueError, match

    def data_scorer_none(self):
        params = {"scorer": None}
        match = self._msg_scorer.format(value=None)
        return params, TypeError, match

    def data_scorer_not_callable(self):
        params = {"scorer": np.zeros(10)}
        value = r"array\(\[0\., 0\., 0\., 0\., 0\., 0\., 0\., 0\., 0\., 0\.\]\)"
        match = self._msg_scorer.format(value=value)
        return params, TypeError, match

    def data_n_estimators_0(self):
        params = {"scorer": score_cindex, "n_estimators": 0}
        match = (
            "The 'n_estimators' parameter of EnsembleSelection must be "
            r"an int in the range \[1, inf\) or a float in the range \(0\.0, 1\.0\]\. "
            r"Got 0 instead\."
        )
        return params, ValueError, match

    def data_n_estimators_1000(self):
        params = {"scorer": score_cindex, "n_estimators": 1000}
        match = r"n_estimators \(1000\) must not exceed number of base learners \(2\)"
        return params, ValueError, match

    def data_correlation_none(self):
        params = {"scorer": score_cindex, "correlation": None}
        match = self._msg_correlation.format(value=None)
        return params, ValueError, match

    def data_correlation_int(self):
        params = {"scorer": score_cindex, "correlation": 2143}
        match = self._msg_correlation.format(value=2143)
        return params, ValueError, match

    def data_correlation_str(self):
        params = {"scorer": score_cindex, "correlation": "clearly wrong"}
        match = self._msg_correlation.format(value="clearly wrong")
        return params, ValueError, match


@pytest.mark.parametrize("params,error_cls,error", EnsembleSelectionFailureCases().get_cases())
@pytest.mark.filterwarnings("ignore:Optimization did not converge.*:sklearn.exceptions.ConvergenceWarning")
def test_ensemble_selection_failures(params, error_cls, error, make_whas500):
    whas500 = make_whas500(with_mean=False, with_std=False, to_numeric=True)
    base_estimators = [
        ("gbm", ComponentwiseGradientBoostingSurvivalAnalysis()),
        ("svm", FastSurvivalSVM()),
    ]
    meta = EnsembleSelection(base_estimators, **params)

    with pytest.raises(error_cls, match=error):
        meta.fit(whas500.x, whas500.y)


def _score_rmse(est, X_test, y_test, **predict_params):
    prediction = est.predict(X_test, **predict_params)

    m = y_test["fstat"]
    res = mean_squared_error(y_test["lenfol"][m], prediction[m])
    return np.sqrt(res)


class DummySurvivalRegressor(DummyRegressor):
    def __init__(self, strategy="mean", constant=None, quantile=None):
        super().__init__(strategy=strategy, constant=constant, quantile=quantile)
        if hasattr(DummyRegressor, "n_features_in_"):
            delattr(DummyRegressor, "n_features_in_")

    def fit(self, X, y, sample_weight=None):
        _, time = check_array_survival(X, y)
        return super().fit(X, time)


def _create_regression_ensemble():
    aft_grid = ParameterGrid({"alpha": 2.0 ** np.arange(-2, 12, 2)})
    svm_grid = ParameterGrid({"alpha": 2.0 ** np.arange(-12, 0, 2)})

    base_estimators = []
    for i, params in enumerate(aft_grid):
        model = IPCRidge(max_iter=1000, **params)
        base_estimators.append((f"aft_{i}", model))

    for i, params in enumerate(svm_grid):
        model = FastSurvivalSVM(rank_ratio=0, fit_intercept=True, max_iter=100, random_state=1, **params)
        base_estimators.append((f"svm_{i}", model))

    cv = KFold(n_splits=4, shuffle=True, random_state=0)
    meta = EnsembleSelectionRegressor(
        base_estimators,
        n_estimators=0.4,
        scorer=_score_rmse,
        cv=cv,
        n_jobs=1,
    )
    return meta


class TestEnsembleSelectionRegressor:
    @staticmethod
    @pytest.mark.slow()
    def test_fit_and_predict(make_whas500):
        whas500 = make_whas500(with_mean=True, with_std=True, to_numeric=True)
        meta = _create_regression_ensemble()
        assert len(meta) == 0

        meta.fit(whas500.x_data_frame.iloc[:400], whas500.y[:400])
        assert meta.scores_.shape[0] >= len(meta)
        assert len(meta) == 5
        assert meta.scores_.shape[0] == 9

        feature_names = whas500.x_data_frame.columns.values
        assert meta.n_features_in_ == len(feature_names)
        assert_array_equal(meta.feature_names_in_, feature_names)

        p = meta.predict(whas500.x_data_frame.iloc[400:])
        score = np.sqrt(mean_squared_error(whas500.y[400:]["lenfol"], p))
        assert abs(score - 423.82894756865056) <= 0.1

        c_index = meta.score(whas500.x_data_frame.iloc[:400], whas500.y[:400])
        assert c_index == pytest.approx(0.767067946974157)

    @staticmethod
    def test_fit_dummy(make_whas500):
        whas500 = make_whas500(with_mean=False, with_std=False, to_numeric=True)
        base_estimators = [
            ("dummy_0", DummySurvivalRegressor(strategy="mean")),
            ("dummy_1", DummySurvivalRegressor(strategy="median")),
            ("dummy_2", DummySurvivalRegressor(strategy="quantile", quantile=0.1)),
            ("dummy_3", DummySurvivalRegressor(strategy="quantile", quantile=0.9)),
            ("dummy_4", DummySurvivalRegressor(strategy="quantile", quantile=0.89)),
            ("dummy_5", DummySurvivalRegressor(strategy="quantile", quantile=0.91)),
        ]

        meta = EnsembleSelectionRegressor(
            base_estimators,
            n_estimators=1,
            min_score=5,
            cv=5,
            scorer=_score_rmse,
        )

        with pytest.raises(ValueError, match="no base estimator exceeds min_score, try decreasing it"):
            meta.fit(whas500.x, whas500.y)

    @staticmethod
    def test_invalid_scorer(make_whas500):
        whas500 = make_whas500(with_mean=False, with_std=False, to_numeric=True)
        base_estimators = [
            ("dummy_0", DummySurvivalRegressor(strategy="mean")),
            ("dummy_1", DummySurvivalRegressor(strategy="median")),
        ]

        def _score(est, X_test, y_test, **predict_params):  # pylint: disable=unused-argument
            return "invalid"

        meta = EnsembleSelectionRegressor(
            base_estimators,
            n_estimators=1,
            min_score=5,
            cv=5,
            scorer=_score,
        )

        with pytest.raises(ValueError, match=r"scoring must return a number, got invalid \(<class 'str'>\) instead\."):
            meta.fit(whas500.x, whas500.y)



from collections import OrderedDict
from io import StringIO

import numpy as np
import pandas as pd
import pandas.testing as tm
import pytest

from survivalist.io import loadarff, writearff
from survivalist.testing import FixtureParameterFactory

EXPECTED_1 = [
    "@relation test_nominal\n",
    "\n",
    "@attribute attr_nominal\t{beer,water,wine}\n",
    '@attribute attr_nominal_spaces\t{"hard liquor",mate,"red wine"}\n',
    "\n",
    "@data\n",
    'water,"red wine"\n',
    'wine,"hard liquor"\n',
    "beer,?\n",
    "?,mate\n",
    'wine,"hard liquor"\n',
    "water,mate\n",
]


EXPECTED_NO_QUOTES = [
    "@relation test_nominal\n",
    "\n",
    "@attribute attr_nominal\t{beer,water,wine}\n",
    "@attribute attr_nominal_spaces\t{hard liquor,mate,red wine}\n",
    "\n",
    "@data\n",
    "water,red wine\n",
    "wine,hard liquor\n",
    "beer,?\n",
    "?,mate\n",
    "wine,hard liquor\n",
    "water,mate\n",
]


EXPECTED_DATETIME = [
    "@relation test_datetime\n",
    "\n",
    "@attribute attr_datetime\tdate 'yyyy-MM-dd HH:mm:ss'\n",
    "\n",
    "@data\n",
    '"2014-10-31 14:13:01"\n',
    '"2004-03-13 19:49:31"\n',
    '"1998-12-06 09:10:11"\n',
]


class DataFrameCases(FixtureParameterFactory):
    def data_nominal(self):
        data = pd.DataFrame(
            {
                "attr_nominal": ["water", "wine", "beer", None, "wine", "water"],
                "attr_nominal_spaces": ["red wine", "hard liquor", None, "mate", "hard liquor", "mate"],
            }
        )
        return data, "test_nominal", EXPECTED_1.copy()

    def data_nominal_with_quotes(self):
        data, rel_name, expected = self.data_nominal()
        data["attr_nominal_spaces"] = ["'red wine'", "'hard liquor'", None, "mate", "'hard liquor'", "mate"]
        return data, rel_name, expected

    def data_nominal_as_category(self):
        data, rel_name, expected = self.data_nominal_with_quotes()
        for name, series in data.items():
            data[name] = pd.Categorical(series, ordered=False)

        expected[3] = '@attribute attr_nominal_spaces\t{"hard liquor","red wine",mate}\n'
        return data, rel_name, expected

    def data_nominal_as_category_extra(self):
        data, rel_name, expected = self.data_nominal_as_category()
        data["attr_nominal"] = pd.Categorical(
            ["water", "wine", "beer", None, "wine", "water"],
            categories=["beer", "coke", "water", "wine"],
            ordered=False,
        )

        expected[2] = "@attribute attr_nominal\t{beer,coke,water,wine}\n"
        return data, rel_name, expected

    def data_nominal_with_category_ordering(self):
        data, rel_name, expected = self.data_nominal_with_quotes()
        data["attr_nominal"] = pd.Categorical(
            ["water", "wine", "beer", None, "wine", "water"],
            categories=["water", "coke", "beer", "wine"],
            ordered=False,
        )

        expected[2] = "@attribute attr_nominal\t{water,coke,beer,wine}\n"
        return data, rel_name, expected

    def data_datetime(self):
        data = pd.DataFrame(
            {
                "attr_datetime": np.array(
                    ["2014-10-31 14:13:01", "2004-03-13 19:49:31", "1998-12-06 09:10:11"], dtype="datetime64"
                )
            }
        )
        return data, "test_datetime", EXPECTED_DATETIME.copy()


def test_loadarff_dataframe():
    contents = "".join(EXPECTED_NO_QUOTES)
    with StringIO(contents) as fp:
        actual_df = loadarff(fp)

    expected_df = pd.DataFrame.from_dict(
        OrderedDict(
            [
                ("attr_nominal", pd.Series(pd.Categorical.from_codes([1, 2, 0, -1, 2, 1], ["beer", "water", "wine"]))),
                (
                    "attr_nominal_spaces",
                    pd.Series(pd.Categorical.from_codes([2, 0, -1, 1, 0, 1], ["hard liquor", "mate", "red wine"])),
                ),
            ]
        )
    )

    tm.assert_frame_equal(expected_df, actual_df, check_exact=True)


@pytest.mark.parametrize("data_frame,relation_name,expectation", DataFrameCases().get_cases())
def test_writearff(data_frame, relation_name, expectation, temp_file):
    writearff(data_frame, temp_file, relation_name=relation_name, index=False)

    with open(temp_file.name) as fp:
        read_date = fp.readlines()

    assert expectation == read_date


def test_writearff_unsupported_column_type(temp_file):
    data = pd.DataFrame(
        {
            "attr_datetime": np.array([2 + 3j, 45.1 - 1j, 0 - 1j, 7 + 0j, 132 - 3j, 1 - 0.41j], dtype="complex128"),
        }
    )

    with pytest.raises(TypeError, match="unsupported type complex128"):
        writearff(data, temp_file, relation_name="test_delta", index=False)



from contextlib import ExitStack
from contextlib import nullcontext as does_not_raise
from io import StringIO
from pathlib import Path
import tempfile

import numpy as np
from numpy.testing import assert_array_almost_equal, assert_array_equal
import pandas as pd
import pandas.testing as tm
import pytest

import survivalist.datasets as sdata
from survivalist.io import writearff
from survivalist.testing import FixtureParameterFactory

ARFF_CATEGORICAL_INDEX_1 = """@relation arff_categorical_index
@attribute index {SampleOne,SampleTwo,SampleThree,SampleFour}
@attribute value real
@attribute label {no,yes}
@attribute size {small,medium,large}
@data
SampleOne,15.1,yes,medium
SampleTwo,13.8,no,large
SampleThree,-0.2,yes,small
SampleFour,2.453,yes,large
"""

ARFF_CATEGORICAL_INDEX_2 = """@relation arff_categorical_index
@attribute index {ASampleOne,ASampleTwo,ASampleThree,ASampleFour,ASampleFive}
@attribute value real
@attribute label {yes,no}
@attribute size {small,medium,large}
@data
ASampleOne,1.51,no,small
ASampleTwo,1.38,no,small
ASampleThree,-20,yes,large
ASampleFour,245.3,yes,small
ASampleFive,3.14,no,large
"""


class Skip:
    pass


@pytest.fixture()
def temp_file_pair():
    tmp_train = tempfile.NamedTemporaryFile("w", suffix=".arff", delete=False)
    tmp_test = tempfile.NamedTemporaryFile("w", suffix=".arff", delete=False)

    yield tmp_train, tmp_test

    Path(tmp_train.name).unlink()
    Path(tmp_test.name).unlink()


def _make_features(n_samples, n_features, seed):
    return np.random.RandomState(seed).randn(n_samples, n_features)


def _make_survival_data(n_samples, n_features, seed):
    rnd = np.random.RandomState(seed)

    x = _make_features(n_samples, n_features, seed)
    event = rnd.binomial(1, 0.2, n_samples)
    time = rnd.exponential(25, size=n_samples)
    return x, event, time


def _make_classification_data(n_samples, n_features, n_classes, seed):
    rnd = np.random.RandomState(seed)

    x = _make_features(n_samples, n_features, seed)
    y = rnd.binomial(n_classes - 1, 0.2, 100)
    return x, y


class GetXyCases(FixtureParameterFactory):
    @property
    def survival_data(self):
        return _make_survival_data(100, 10, 0)

    @property
    def features(self):
        return _make_features(100, 10, 0)

    @property
    def attr_labels(self):
        return ["event", "time"]

    def _to_data_frame(self, data, columns):
        if isinstance(
            data,
            (
                tuple,
                list,
            ),
        ):
            data = np.column_stack(data)
        return pd.DataFrame(data, columns=columns)

    @property
    def columns(self):
        return [f"V{i}" for i in range(10)]

    def data_survival_data(self):
        x, event, time = self.survival_data
        attr_labels = self.attr_labels
        dataset = self._to_data_frame((x, event, time), self.columns + attr_labels)

        args = (dataset, attr_labels)
        kwargs = {"pos_label": 1, "survival": True}
        return args, kwargs, x, (event, time), does_not_raise()

    def data_no_label(self):
        x = self.features
        attr_labels = [None, None]
        dataset = self._to_data_frame(x, self.columns)

        args = (dataset, attr_labels)
        kwargs = {"pos_label": 1, "survival": True}
        return args, kwargs, x, None, does_not_raise()

    def data_too_many_labels(self):
        x, event, time = self.survival_data
        attr_labels = self.attr_labels + ["random"]
        dataset = self._to_data_frame((x, event, time), self.columns + self.attr_labels)

        args = (dataset, attr_labels)
        kwargs = {"pos_label": 1, "survival": True}
        error = pytest.raises(
            ValueError,
            match="expected sequence of length two for attr_labels, but got 3",
        )
        return args, kwargs, Skip(), Skip(), error

    def data_too_little_labels_0(self):
        x, event, time = self.survival_data
        attr_labels = self.attr_labels[:1]
        dataset = self._to_data_frame((x, event, time), self.columns + self.attr_labels)

        args = (dataset, attr_labels)
        kwargs = {"pos_label": 1, "survival": True}
        error = pytest.raises(
            ValueError,
            match="expected sequence of length two for attr_labels, but got 1",
        )
        return args, kwargs, Skip(), Skip(), error

    def data_too_little_labels_1(self):
        x, event, time = self.survival_data
        attr_labels = []
        dataset = self._to_data_frame((x, event, time), self.columns + self.attr_labels)

        args = (dataset, attr_labels)
        kwargs = {"pos_label": 1, "survival": True}
        error = pytest.raises(
            ValueError,
            match="expected sequence of length two for attr_labels, but got 0",
        )
        return args, kwargs, Skip(), Skip(), error

    def data_no_pos_label(self):
        x, event, time = self.survival_data
        attr_labels = self.attr_labels
        dataset = self._to_data_frame((x, event, time), self.columns + attr_labels)

        args = (dataset, attr_labels)
        kwargs = {"survival": True}
        error = pytest.raises(
            ValueError,
            match="pos_label needs to be specified if survival=True",
        )
        return args, kwargs, Skip(), Skip(), error

    def data_classification(self):
        x, label = _make_classification_data(100, 10, 6, 0)
        attr_labels = ["class_label"]
        dataset = self._to_data_frame((x, label), self.columns + attr_labels)

        args = (dataset, attr_labels)
        kwargs = {"survival": False}
        return args, kwargs, x, label, does_not_raise()

    def data_classification_no_label(self):
        x = self.features
        attr_labels = None
        dataset = self._to_data_frame(x, self.columns)

        args = (dataset, attr_labels)
        kwargs = {"survival": False}
        return args, kwargs, x, None, does_not_raise()


@pytest.mark.parametrize("args,kwargs,x_expected,y_expected,error_expected", GetXyCases().get_cases())
def test_get_xy(args, kwargs, x_expected, y_expected, error_expected):
    with error_expected:
        x_test, y_test = sdata.get_x_y(*args, **kwargs)

    if not isinstance(x_expected, Skip):
        assert_array_equal(x_test, x_expected)

    if not isinstance(y_expected, Skip):
        if y_expected is None:
            assert y_test is None
        elif isinstance(y_expected, tuple):
            assert y_test.dtype.names == ("event", "time")
            event, time = y_expected
            assert_array_equal(y_test["event"].astype(np.uint32), event.astype(np.uint32))
            assert_array_almost_equal(y_test["time"], time)
        else:
            assert y_test.ndim == 2
            assert_array_equal(y_test.values.ravel(), y_expected)


def assert_structured_array_dtype(arr, event, time, num_events):
    assert arr.dtype.names == (event, time)
    assert np.issubdtype(arr.dtype.fields[event][0], np.bool_)
    assert np.issubdtype(arr.dtype.fields[time][0], np.float64)
    assert arr[event].sum() == num_events


class TestLoadDatasets:
    @staticmethod
    def test_load_whas500():
        x, y = sdata.load_whas500()
        assert x.shape == (500, 14)
        assert y.shape == (500,)
        assert_structured_array_dtype(y, "fstat", "lenfol", 215)

    @staticmethod
    def test_load_gbsg2():
        x, y = sdata.load_gbsg2()
        assert x.shape == (686, 8)
        assert y.shape == (686,)
        assert_structured_array_dtype(y, "cens", "time", 299)

    @staticmethod
    def test_load_veterans_lung_cancer():
        x, y = sdata.load_veterans_lung_cancer()
        assert x.shape == (137, 6)
        assert y.shape == (137,)
        assert_structured_array_dtype(y, "Status", "Survival_in_days", 128)

    @staticmethod
    def test_load_aids():
        x, y = sdata.load_aids(endpoint="aids")
        assert x.shape == (1151, 11)
        assert y.shape == (1151,)
        assert_structured_array_dtype(y, "censor", "time", 96)
        assert "censor_d" not in x.columns
        assert "time_d" not in x.columns

        x, y = sdata.load_aids(endpoint="death")
        assert x.shape == (1151, 11)
        assert y.shape == (1151,)
        assert_structured_array_dtype(y, "censor_d", "time_d", 26)
        assert "censor" not in x.columns
        assert "time" not in x.columns

        with pytest.raises(ValueError, match="endpoint must be 'aids' or 'death'"):
            sdata.load_aids(endpoint="foobar")

    @staticmethod
    def test_load_breast_cancer():
        x, y = sdata.load_breast_cancer()
        assert x.shape == (198, 80)
        assert y.shape == (198,)
        assert_structured_array_dtype(y, "e.tdm", "t.tdm", 51)

    @staticmethod
    def test_load_flchain():
        x, y = sdata.load_flchain()
        assert x.shape == (7874, 9)
        assert y.shape == (7874,)
        assert_structured_array_dtype(y, "death", "futime", 2169)


def _make_and_write_data(fp, n_samples, n_features, with_index, with_labels, seed, column_prefix="V"):
    x, event, time = _make_survival_data(n_samples, n_features, seed)

    columns = [f"{column_prefix}{i}" for i in range(n_features)]
    if with_labels:
        columns += ["event", "time"]
        arr = np.column_stack((x, event, time))
    else:
        arr = x

    if with_index:
        index = np.arange(n_samples, dtype=float)
        np.random.RandomState(0).shuffle(index)
    else:
        index = None

    dataset = pd.DataFrame(arr, index=index, columns=columns)
    dataset.index.name = "index"

    writearff(dataset, fp, index=with_index)
    return dataset


def assert_x_equal(x_true, x_train):
    tm.assert_index_equal(x_true.columns, x_train.columns, exact=True)
    tm.assert_index_equal(x_true.index, x_train.index, exact=True)

    tm.assert_frame_equal(
        x_true,
        x_train,
        check_index_type=False,
        check_column_type=True,
        check_names=False,
    )


def assert_y_equal(y_true, y_train):
    assert y_train.dtype.names == ("event", "time")

    assert_array_equal(
        y_train["event"].astype(np.uint32),
        y_true["event"].values.astype(np.uint32),
    )
    assert_array_almost_equal(y_train["time"], y_true["time"].values)


class LoadArffFilesCases(FixtureParameterFactory):
    @property
    def arff_1(self):
        return StringIO(ARFF_CATEGORICAL_INDEX_1)

    @property
    def arff_2(self):
        return StringIO(ARFF_CATEGORICAL_INDEX_2)

    def data_with_categorical_index_1(self):
        values = ["SampleOne", "SampleTwo", "SampleThree", "SampleFour"]
        index = pd.Index(values, name="index", dtype=object)
        x = pd.DataFrame.from_dict(
            {
                "size": pd.Series(
                    pd.Categorical(
                        ["medium", "large", "small", "large"],
                        categories=["small", "medium", "large"],
                        ordered=False,
                    ),
                    name="size",
                ),
                "value": pd.Series([15.1, 13.8, -0.2, 2.453], name="value"),
            }
        )
        x.index = index

        y = pd.DataFrame.from_dict(
            {
                "label": pd.Series(
                    pd.Categorical(["yes", "no", "yes", "yes"], categories=["no", "yes"], ordered=False), name="label"
                )
            }
        )
        y.index = index

        args = (self.arff_1, ["label"])
        kwargs = {
            "pos_label": "yes",
            "survival": False,
            "standardize_numeric": False,
            "to_numeric": False,
        }

        return args, kwargs, x, y, None, None

    def data_with_categorical_index_2(self):
        values = ["ASampleOne", "ASampleTwo", "ASampleThree", "ASampleFour", "ASampleFive"]
        index = pd.Index(values, name="index", dtype=object)

        y = pd.DataFrame.from_dict(
            {
                "label": pd.Series(
                    pd.Categorical(["no", "no", "yes", "yes", "no"], categories=["yes", "no"], ordered=False),
                    name="label",
                )
            }
        )
        y.index = index

        x = pd.DataFrame.from_dict(
            {
                "size": pd.Series(
                    pd.Categorical(
                        ["small", "small", "large", "small", "large"],
                        categories=["small", "medium", "large"],
                        ordered=False,
                    ),
                    name="size",
                ),
                "value": pd.Series([1.51, 1.38, -20, 245.3, 3.14], name="value"),
            }
        )
        x.index = index

        args = (self.arff_2, ["label"])
        kwargs = {
            "pos_label": "yes",
            "survival": False,
            "standardize_numeric": False,
            "to_numeric": False,
        }

        return args, kwargs, x, y, None, None

    def data_with_categorical_index(self):
        _, _, x_train, y_train, _, _ = self.data_with_categorical_index_1()
        _, _, x_test, y_test, _, _ = self.data_with_categorical_index_2()

        args = (self.arff_1, ["label"])
        kwargs = {
            "pos_label": "yes",
            "path_testing": self.arff_2,
            "survival": False,
            "standardize_numeric": False,
            "to_numeric": False,
        }

        return args, kwargs, x_train, y_train, x_test, y_test


@pytest.mark.parametrize(
    "args,kwargs,x_train_expected,y_train_expected,x_test_expected,y_test_expected",
    LoadArffFilesCases().get_cases(),
)
def test_load_arff_files(
    args,
    kwargs,
    x_train_expected,
    y_train_expected,
    x_test_expected,
    y_test_expected,
):
    x_train, y_train, x_test, y_test = sdata.load_arff_files_standardized(
        *args,
        **kwargs,
    )

    tm.assert_frame_equal(x_train, x_train_expected, check_exact=True)
    tm.assert_frame_equal(y_train, y_train_expected, check_exact=True)

    if x_test_expected is None:
        assert x_test is None
    else:
        tm.assert_frame_equal(x_test, x_test_expected, check_exact=True)

    if y_test_expected is None:
        assert y_test is None
    else:
        tm.assert_frame_equal(y_test, y_test_expected, check_exact=True)


class LoadArffFilesWithTempFileCases(FixtureParameterFactory):
    def data_with_index(self):
        args_train = (100, 10, True, True, 0)
        args_test = None
        return args_train, {}, args_test, {}, (does_not_raise(),)

    def data_train_and_test_with_labels(self):
        args_train = (100, 10, True, True, 0)
        args_test = (20, 10, True, True, 0)
        return args_train, {}, args_test, {}, (does_not_raise(),)

    def data_train_and_test_no_labels(self):
        args_train = (100, 10, True, True, 0)
        args_test = (20, 10, True, False, 0)
        return args_train, {}, args_test, {}, (does_not_raise(),)

    def data_train_and_test_with_different_columns(self):
        args_train = (100, 19, False, True, 0)
        args_test = (20, 11, False, True, 0)
        error = pytest.warns(
            UserWarning,
            match="Restricting columns to intersection between training and testing data",
        )
        return args_train, {}, args_test, {}, (error,)

    def data_train_and_test_columns_dont_intersect(self):
        args_train = (100, 19, True, True, 0)
        kwargs_train = {"column_prefix": "A"}
        args_test = (20, 11, True, True, 0)
        kwargs_test = {"column_prefix": "B"}
        error = pytest.raises(
            ValueError,
            match="columns of training and test data do not intersect",
        )
        warning = pytest.warns(
            UserWarning,
            match="Restricting columns to intersection between training and testing data",
        )
        return (
            args_train,
            kwargs_train,
            args_test,
            kwargs_test,
            (
                error,
                warning,
            ),
        )


@pytest.mark.parametrize(
    "args_train,kwargs_train,args_test,kwargs_test,errors_expected", LoadArffFilesWithTempFileCases().get_cases()
)
def test_load_from_temp_file(args_train, kwargs_train, args_test, kwargs_test, errors_expected, temp_file_pair):
    tmp_train, tmp_test = temp_file_pair

    train_dataset = _make_and_write_data(tmp_train, *args_train, **kwargs_train)
    if args_test is not None:
        test_dataset = _make_and_write_data(tmp_test, *args_test, **kwargs_test)
        path_testing = tmp_test.name
        check_y_test = args_test[-2]  # with_label
    else:
        test_dataset = None
        path_testing = None
        check_y_test = False
        tmp_test.close()

    with ExitStack() as stack:
        for error_expected in errors_expected:
            stack.enter_context(error_expected)
        x_train, y_train, x_test, y_test = sdata.load_arff_files_standardized(
            tmp_train.name,
            ["event", "time"],
            1,
            path_testing=path_testing,
            survival=True,
            standardize_numeric=False,
            to_numeric=False,
        )

    if all(not isinstance(err, does_not_raise) for err in errors_expected):
        return

    cols = ["event", "time"]

    x_true = train_dataset.drop(cols, axis=1)
    assert_x_equal(x_true, x_train)
    assert_y_equal(train_dataset, y_train)

    if test_dataset is not None:
        x_true = test_dataset
        if check_y_test:
            assert_y_equal(test_dataset, y_test)
            x_true = test_dataset.drop(cols, axis=1)
        else:
            assert y_test is None
        assert_x_equal(x_true, x_test)
    else:
        assert x_test is None
        assert y_test is None



import numpy as np
from numpy.testing import assert_array_almost_equal
import pytest

from survivalist.linear_model import CoxnetSurvivalAnalysis
from survivalist.testing import all_survival_estimators
from survivalist.util import Surv


def all_survival_function_estimators():
    estimators = set()
    for cls in all_survival_estimators():
        if hasattr(cls, "predict_survival_function"):
            if issubclass(cls, CoxnetSurvivalAnalysis):
                est = cls(fit_baseline_model=True)
            else:
                est = cls()
            estimators.add(est)
    return estimators


@pytest.mark.parametrize("estimator", all_survival_function_estimators())
def test_survival_functions(estimator, make_whas500):
    data = make_whas500(to_numeric=True)

    estimator.fit(data.x[150:], data.y[150:])
    fns_cls = estimator.predict_survival_function(data.x[:150])
    fns_arr = estimator.predict_survival_function(data.x[:150], return_array=True)

    times = estimator.unique_times_
    arr = np.vstack([fn(times) for fn in fns_cls])

    assert_array_almost_equal(arr, fns_arr)


@pytest.mark.parametrize("estimator", all_survival_function_estimators())
@pytest.mark.parametrize("y_time", [-1e-8, -1, np.finfo(float).min])
def test_fit_negative_survial_time_raises(estimator, y_time):
    X = np.random.randn(7, 3)
    y = Surv.from_arrays(event=np.ones(7, dtype=bool), time=[1, 9, 3, y_time, 1, 8, 1e10])

    with pytest.raises(ValueError, match="observed time contains values smaller zero"):
        estimator.fit(X, y)



import numpy as np
from numpy.testing import assert_array_almost_equal
import pytest
from sklearn.pipeline import make_pipeline

from survivalist.base import SurvivalAnalysisMixin
from survivalist.linear_model import IPCRidge
from survivalist.testing import assert_cindex_almost_equal


class TestIPCRidge:
    @staticmethod
    def test_fit(make_whas500):
        whas500 = make_whas500()
        model = IPCRidge()
        model.fit(whas500.x, whas500.y)

        assert model.intercept_ == pytest.approx(5.867520370855396, 1e-7)
        expected = np.array(
            [
                0.168481,
                -0.24962,
                2.185086,
                0.53682,
                -0.514611,
                0.09124,
                0.613114,
                0.480357,
                -0.055972,
                0.238472,
                -0.127209,
                -0.144063,
                -1.625081,
                -0.217591,
            ]
        )
        assert_array_almost_equal(model.coef_, expected)

    @staticmethod
    def test_predict(make_whas500):
        whas500 = make_whas500()
        model = IPCRidge()
        model.fit(whas500.x[:400], whas500.y[:400])

        x_test = whas500.x[400:]
        y_test = whas500.y[400:]
        p = model.predict(x_test)
        assert_cindex_almost_equal(
            y_test["fstat"],
            y_test["lenfol"],
            -p,
            (0.66925817946226107, 2066, 1021, 0, 1),
        )

        assert model.score(x_test, y_test) == 0.66925817946226107

    @staticmethod
    def test_pipeline_score(make_whas500):
        whas500 = make_whas500()
        pipe = make_pipeline(IPCRidge())
        pipe.fit(whas500.x[:400], whas500.y[:400])

        x_test = whas500.x[400:]
        y_test = whas500.y[400:]
        p = pipe.predict(x_test)
        assert_cindex_almost_equal(
            y_test["fstat"],
            y_test["lenfol"],
            -p,
            (0.66925817946226107, 2066, 1021, 0, 1),
        )

        assert SurvivalAnalysisMixin.score(pipe, x_test, y_test) == 0.66925817946226107



import numpy as np

from . import functional as ssf


class SurvivalStacker:
    """Casts a survival analysis problem as a classification problem as
    proposed in Craig E., et al. 2021 (arXiv:2107.13480)
    """

    def __init__(self, times: np.ndarray | None = None) -> None:
        """Generate a SurvivalStacker instance

        :param times: array of time points on which to create risk sets
        """
        if times is not None:
            self.times = times
        else:
            self.times = np.empty(0)

    def fit(self, X: np.ndarray, y: np.ndarray, time_step: float | None = None):
        """Generate the risk time points

        :param X: survival input samples
        :param y: structured array with two fields. The binary event indicator
            as first field, and time of event or time of censoring as
            second field.
        :param time_step: a base multiple on which to bin times. If none, the
            times for all observed events are used.
        :return: self
        """
        event_field, time_field = y.dtype.names
        event_times = np.unique(y[time_field][y[event_field]])
        if time_step is None:
            self.times = event_times
        else:
            self.times = ssf.digitize_times(event_times, time_step)
        return self

    def transform(self, X: np.ndarray, y: np.ndarray | None = None) -> tuple[np.ndarray, np.ndarray | None]:
        """Convert the input survival dataset to a stacked survival dataset

        :param X: survival input samples
        :param y: structured array with two fields. The binary event indicator
            as first field, and time of event or time of censoring as
            second field. If None, the returned dataset is constructed for
            evaluation.
        :return: a tuple containing the predictor matrix and response vector
        """
        if y is None:
            X_stacked = ssf.stack_eval(X, self.times)
            y_stacked = None
        else:
            X_stacked, y_stacked = ssf.stack_timepoints(X, y, self.times)
        return X_stacked, y_stacked

    def fit_transform(
        self, X: np.ndarray, y: np.ndarray, time_step: float | None = None
    ) -> tuple[np.ndarray, np.ndarray | None]:
        """Fit to data, then transform it.

        :param X: survival input samples
        :param y: structured array with two fields. The binary event indicator
            as first field, and time of event or time of censoring as
            second field.
        :param time_step: a base multiple on which to bin times. If none, the
            times for all observed events are used.
        :return: a tuple containing the predictor matrix and response vector
        """
        self.fit(X, y, time_step)
        return self.transform(X, y)

    def cumulative_hazard_function(self, X: np.ndarray) -> np.ndarray:
        """Calculate the cumulative hazard function from the stacked survival
        estimates.

        :param X: estimates as returned from a model trained on
        an evaluation set
        :return: a cumulative risk matrix for the fitted time-points
        """
        return ssf.cumulative_hazard_function(X, self.times)

    def predict_survival_function(self, X: np.ndarray) -> np.ndarray:
        """Calculate the survival function from the stacked survival estimates.

        :param X: estimates as returned from a model trained on
        an evaluation set
        :return: the survival function for the fitted time-points
        """
        return ssf.survival_function(X, self.times)

    def risk_score(self, estimates: np.ndarray) -> np.ndarray:
        """Calculate risk score from stacked survival estimates.

        :param estimates: estimates as returned from a model trained on
        an evaluation set
        :return: the risk score
        """
        return ssf.risk_score(estimates, self.times)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging

import numpy as np
import pandas as pd
from pandas.api.types import CategoricalDtype, is_object_dtype

__all__ = ["categorical_to_numeric", "encode_categorical", "standardize"]


def _apply_along_column(array, func1d, **kwargs):
    if isinstance(array, pd.DataFrame):
        return array.apply(func1d, **kwargs)
    return np.apply_along_axis(func1d, 0, array, **kwargs)


def standardize_column(series_or_array, with_std=True):
    d = series_or_array.dtype
    if issubclass(d.type, np.number):
        output = series_or_array.astype(float)
        m = series_or_array.mean()
        output -= m

        if with_std:
            s = series_or_array.std(ddof=1)
            output /= s

        return output

    return series_or_array


def standardize(table, with_std=True):
    """
    Perform Z-Normalization on each numeric column of the given table.

    If `table` is a pandas.DataFrame, only numeric columns are modified,
    all other columns remain unchanged. If `table` is a numpy.ndarray,
    it is only modified if it has numeric dtype, in which case the returned
    array will have floating point dtype.

    Parameters
    ----------
    table : pandas.DataFrame or numpy.ndarray
        Data to standardize.

    with_std : bool, optional, default: True
        If ``False`` data is only centered and not converted to unit variance.

    Returns
    -------
    normalized : pandas.DataFrame
        Table with numeric columns normalized.
        Categorical columns in the input table remain unchanged.
    """
    new_frame = _apply_along_column(
        table, standardize_column, with_std=with_std)

    return new_frame


def _encode_categorical_series(series, allow_drop=True):
    values = _get_dummies_1d(series, allow_drop=allow_drop)
    if values is None:
        return

    enc, levels = values
    if enc is None:
        return pd.Series(index=series.index, name=series.name, dtype=series.dtype)

    if not allow_drop and enc.shape[1] == 1:
        return series

    names = []
    for key in range(1, enc.shape[1]):
        names.append(f"{series.name}={levels[key]}")
    series = pd.DataFrame(enc[:, 1:], columns=names, index=series.index)

    return series


def encode_categorical(table, columns=None, **kwargs):
    """
    Encode categorical columns with `M` categories into `M-1` columns according
    to the one-hot scheme.

    Parameters
    ----------
    table : pandas.DataFrame
        Table with categorical columns to encode.

    columns : list-like, optional, default: None
        Column names in the DataFrame to be encoded.
        If `columns` is None then all the columns with
        `object` or `category` dtype will be converted.

    allow_drop : boolean, optional, default: True
        Whether to allow dropping categorical columns that only consist
        of a single category.

    Returns
    -------
    encoded : pandas.DataFrame
        Table with categorical columns encoded as numeric.
        Numeric columns in the input table remain unchanged.
    """
    if isinstance(table, pd.Series):
        if not isinstance(table.dtype, CategoricalDtype) and not is_object_dtype(table.dtype):
            raise TypeError(
                f"series must be of categorical dtype, but was {table.dtype}")
        return _encode_categorical_series(table, **kwargs)

    def _is_categorical_or_object(series):
        return isinstance(series.dtype, CategoricalDtype) or is_object_dtype(series.dtype)

    if columns is None:
        # for columns containing categories
        columns_to_encode = {nam for nam,
                             s in table.items() if _is_categorical_or_object(s)}
    else:
        columns_to_encode = set(columns)

    items = []
    for name, series in table.items():
        if name in columns_to_encode:
            series = _encode_categorical_series(series, **kwargs)
            if series is None:
                continue
        items.append(series)

    # concat columns of tables
    new_table = pd.concat(items, axis=1, copy=False)
    return new_table


def _get_dummies_1d(data, allow_drop=True):
    # Series avoids inconsistent NaN handling
    cat = pd.Categorical(data)
    levels = cat.categories
    number_of_cols = len(levels)

    # if all NaN or only one level
    if allow_drop and number_of_cols < 2:
        logging.getLogger(__package__).warning(
            f"dropped categorical variable {data.name!r}, because it has only {number_of_cols} values"
        )
        return
    if number_of_cols == 0:
        return None, levels

    dummy_mat = np.eye(number_of_cols).take(cat.codes, axis=0)

    # reset NaN GH4446
    dummy_mat[cat.codes == -1] = np.nan

    return dummy_mat, levels


def categorical_to_numeric(table):
    """Encode categorical columns to numeric by converting each category to
    an integer value.

    Parameters
    ----------
    table : pandas.DataFrame
        Table with categorical columns to encode.

    Returns
    -------
    encoded : pandas.DataFrame
        Table with categorical columns encoded as numeric.
        Numeric columns in the input table remain unchanged.
    """

    def transform(column):
        if isinstance(column.dtype, CategoricalDtype):
            return column.cat.codes
        if is_object_dtype(column.dtype):
            try:
                nc = column.astype(np.int64)
            except ValueError:
                classes = column.dropna().unique()
                classes.sort(kind="mergesort")
                nc = column.map(dict(zip(classes, range(classes.shape[0]))))
            return nc
        if column.dtype == bool:
            return column.astype(np.int64)

        return column

    if isinstance(table, pd.Series):
        return pd.Series(transform(table), name=table.name, index=table.index)
    return table.apply(transform, axis=0, result_type="expand")



from importlib.metadata import PackageNotFoundError, version
import platform
import sys

from sklearn.pipeline import Pipeline, _final_estimator_has
from sklearn.utils.metaestimators import available_if

from .util import property_available_if


def _get_version(name):
    try:
        pkg_version = version(name)
    except ImportError:
        pkg_version = None
    return pkg_version


def show_versions():
    sys_info = {
        "Platform": platform.platform(),
        "Python version": f"{platform.python_implementation()} {platform}",
        "Python interpreter": sys.executable,
    }

    deps = [
        "survivalist",
        "scikit-learn",
        "numpy",
        "scipy",
        "pandas",
        "numexpr",
        "ecos",
        "osqp",
        "joblib",
        "matplotlib",
        "pytest",
        "sphinx",
        "Cython",
        "pip",
        "setuptools",
    ]
    minwidth = max(
        max(map(len, deps)),
        max(map(len, sys_info.keys())),
    )
    fmt = "{0:<%ds}: {1}" % minwidth

    print("SYSTEM")
    print("------")
    for name, version_string in sys_info.items():
        print(fmt.format(name, version_string))

    print("\nDEPENDENCIES")
    print("------------")
    for dep in deps:
        version_string = _get_version(dep)
        print(fmt.format(dep, version_string))


@available_if(_final_estimator_has("predict_cumulative_hazard_function"))
def predict_cumulative_hazard_function(self, X, **kwargs):
    """Predict cumulative hazard function.

    The cumulative hazard function for an individual
    with feature vector :math:`x` is defined as

    .. math::

        H(t \\mid x) = \\exp(x^\\top \\beta) H_0(t) ,

    where :math:`H_0(t)` is the baseline hazard function,
    estimated by Breslow's estimator.

    Parameters
    ----------
    X : array-like, shape = (n_samples, n_features)
        Data matrix.

    Returns
    -------
    cum_hazard : ndarray, shape = (n_samples,)
        Predicted cumulative hazard functions.
    """
    Xt = X
    for _, _, transform in self._iter(with_final=False):
        Xt = transform.transform(Xt)
    return self.steps[-1][-1].predict_cumulative_hazard_function(Xt, **kwargs)


@available_if(_final_estimator_has("predict_survival_function"))
def predict_survival_function(self, X, **kwargs):
    """Predict survival function.

    The survival function for an individual
    with feature vector :math:`x` is defined as

    .. math::

        S(t \\mid x) = S_0(t)^{\\exp(x^\\top \\beta)} ,

    where :math:`S_0(t)` is the baseline survival function,
    estimated by Breslow's estimator.

    Parameters
    ----------
    X : array-like, shape = (n_samples, n_features)
        Data matrix.

    Returns
    -------
    survival : ndarray, shape = (n_samples,)
        Predicted survival functions.
    """
    Xt = X
    for _, _, transform in self._iter(with_final=False):
        Xt = transform.transform(Xt)
    return self.steps[-1][-1].predict_survival_function(Xt, **kwargs)


@property_available_if(_final_estimator_has("_predict_risk_score"))
def _predict_risk_score(self):
    return self.steps[-1][-1]._predict_risk_score


def patch_pipeline():
    Pipeline.predict_survival_function = predict_survival_function
    Pipeline.predict_cumulative_hazard_function = predict_cumulative_hazard_function
    Pipeline._predict_risk_score = _predict_risk_score


try:
    __version__ = version("survivalist")
except PackageNotFoundError:  # pragma: no cover
    # package is not installed
    __version__ = "unknown"

patch_pipeline()
